import unittest
import contextlib
import logging
import re
import datetime as DT
import pytz
import os
import sys
import subprocess
from esahub.tests import config as test_config
from esahub import config
from esahub import geo, helpers, scihub, checksum, check, main

logger = logging.getLogger('esahub')

PY2 = sys.version_info < (3, 0)
SMALL_SIZE_QUERY = 'size: ???.* KB'

if hasattr(unittest.TestCase, 'subTest'):
    class TestCase(unittest.TestCase):
        pass
else:
    class TestCase(unittest.TestCase):
        @contextlib.contextmanager
        def subTest(self, msg='', **params):
            """Mock subTest method so no exception is raised under Python2."""
            helpers.eprint('subTest:', msg, params)
            yield
            return


# -----------------------------------------------------------------------------
# TEST SETUP
# -----------------------------------------------------------------------------
def setUpModule():
    test_config.set_test_config()
    test_config.prepare()


def tearDownModule():
    test_config.cleanup()


# -----------------------------------------------------------------------------
# SCIHUB
# -----------------------------------------------------------------------------
class ScihubTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    # def setUp(self):

    def test_servers(self):
        for name in scihub._get_available_servers():
            auth = config.CONFIG['SERVERS'][name]
            conn = scihub.Connection(**auth)
            with self.subTest(server_name=name):
                # url = scihub_connect.Query().build_query_url()
                url = '{}/search?q=*:*'.format(auth['host'])
                result = conn.resolve(url)
                #
                # Assert that the HTML response has status code 200 (OK)
                #
                self.assertEqual(result.code, 200)

    def test__generate_next_url(self):
        # _generate_next_url(url, total=None)
        pass

    def test__parse_page(self):
        # _parse_page(url, first=False)
        pass

    def test__get_file_list_from_url(self):
        # _get_file_list_from_url(url, limit=None)
        pass

    def test__callback(self):
        # _callback(result)
        pass

    def test__build_query(self):
        # _build_query(query={})
        pass

    def test__build_url(self):
        # _build_url(query, server)
        pass

    def test__download(self):
        # _download(url, destination, quiet=None, queue=None)
        pass

    def test__get_file_list_wrapper(self):
        # _get_file_list_wrapper(url)
        pass

    def test__ping_single(self):
        # _ping_single(servername)
        pass

    def test__auto_detect_server_from_query(self):
        queries = [
            # (query, server)
            ({'mission': 'Sentinel-1'},
             config.CONFIG['SATELLITES']['S1A']['source']),
            ({'mission': 'Sentinel-2'},
             config.CONFIG['SATELLITES']['S2A']['source']),
            ({'mission': 'Sentinel-3'},
             config.CONFIG['SATELLITES']['S3A']['source']),
            ({'satellite': 'S1A'},
             config.CONFIG['SATELLITES']['S1A']['source']),
            ({'satellite': 'S3A'},
             config.CONFIG['SATELLITES']['S3A']['source']),
            ({'satellite': 'S2B'},
             config.CONFIG['SATELLITES']['S2B']['source']),
            ({'identifier': "S1A_IW_OCN__2SDV_20160924T181320_"
                            "20160924T181345_013198_014FDF_6692.zip"},
             config.CONFIG['SATELLITES']['S1A']['source'])
        ]
        for query, server in queries:
            with self.subTest(query=query):
                self.assertEqual(
                    scihub._auto_detect_server_from_query(query), server
                )

    def test__uuid_from_identifier(self):
        products = scihub.search({}, limit=1)
        for product in products:
            with self.subTest(product=product):
                self.assertEqual(
                    scihub._uuid_from_identifier(product['title']),
                    product['uuid']
                )

    # def test__download_url_from_identifier(self):
    #     # _download_url_from_identifier(identifier)
    #     pass
    # def test__checksum_url_from_identifier(self):
    #     # _checksum_url_from_identifier(identifier)
    #     pass
    # def test__preview_url_from_identifier(self):
    #     # _preview_url_from_identifier(identifier)
    #     pass
    # def test__download_url_from_uuid(self):
    #     # _download_url_from_uuid(uuid, host=None)
    #     pass
    # def test__checksum_url_from_uuid(self):
    #     # _checksum_url_from_uuid(uuid, host=None)
    #     pass
    # def test__preview_url_from_uuid(self):
    #     # _preview_url_from_uuid(uuid, host=None)
    #     pass

    def test_resolve(self):
        for name in scihub._get_available_servers():
            with self.subTest(server_name=name):
                response = scihub.resolve(
                    scihub._build_url({'query': '*:*'}, name)
                )
                self.assertEqual(response.code, 200)

    def test_md5_from_file(self):
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the md5 sum computed from the local file is equal
                # to the md5 sum obtained from the remote server.
                #
                try:
                    remote_md5 = scihub.md5(f)
                    self.assertEqual(
                        checksum.md5(f), remote_md5
                    )
                except Exception as e:
                    self.fail('Remote MD5 could not be obtained: {}'.format(e))

    def test_exists_true(self):
        existing = scihub.search({}, limit=1)
        for e in existing:
            with self.subTest(product=e['filename']):
                self.assertTrue(scihub.exists(e['filename']))

    def test_exists_false(self):
        not_existing = 'this_is_not_on_scihub'
        self.assertFalse(scihub.exists(not_existing))


# -----------------------------------------------------------------------------
# SCIHUB SEARCH
# -----------------------------------------------------------------------------
class ScihubSearchTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def test_query_entries(self):
        query = {'mission': 'Sentinel-3'}
        server = scihub._auto_detect_server_from_query(query,
                                                       available_only=True)[0]
        url = scihub._build_url(query, server)
        result = scihub.resolve(url)
        html = result.read()
        #
        # Assert that the number of entries found on the page matches the
        # number of entries requested per page.
        #
        self.assertEqual(html.count(b'<entry>'),
                         config.CONFIG['GENERAL']['ENTRIES'])

    def test_queries(self):
        queries = [
            # (name, query)
            ('S3', {'mission': 'Sentinel-3'}),
        ]
        for name, q in queries:
            with self.subTest(name=name):
                server = scihub._auto_detect_server_from_query(
                    q, available_only=True)[0]
                url = scihub._build_url(q, server=server)
                result = scihub.resolve(url)
                #
                # Assert that queries for each mission return a
                # status code 200 (OK)
                #
                self.assertEqual(result.code, 200)

        with self.subTest('count entries'):
            q = {'mission': 'Sentinel-3'}
            server = scihub._auto_detect_server_from_query(
                q, available_only=True)[0]
            url = scihub._build_url(q, server=server)
            result = scihub.resolve(url)
            html = result.read()
            #
            # Assert that the number of entries found on the page matches the
            # number of entries requested per page.
            #
            self.assertEqual(html.count(b'<entry>'),
                             config.CONFIG['GENERAL']['ENTRIES'])

    def test_temporal_queries(self):
        with self.subTest('yesterday'):
            file_list = scihub.search({'mission': 'Sentinel-3',
                                       'time': 'yesterday'},
                                      limit=200)
            yesterday = DT.datetime.now(pytz.utc)-DT.timedelta(1)
            today = DT.datetime.now(pytz.utc)
            start = DT.datetime(yesterday.year, yesterday.month, yesterday.day,
                                tzinfo=pytz.utc)
            end = DT.datetime(today.year, today.month, today.day,
                              tzinfo=pytz.utc)
            for f in file_list:
                #
                # Assert that the ingestiondate of each entry was yesterday.
                #
                self.assertGreaterEqual(f['ingestiondate'], start)
                self.assertLessEqual(f['ingestiondate'], end)

        with self.subTest('today'):
            file_list = scihub.search({'mission': 'Sentinel-3',
                                       'time': 'today'},
                                      limit=200)
            today = DT.datetime.now(pytz.utc)
            start = DT.datetime(today.year, today.month, today.day,
                                tzinfo=pytz.utc)
            for f in file_list:
                #
                # Assert that the ingestiondate of each entry is today.
                #
                self.assertGreaterEqual(f['ingestiondate'], start)

    #
    # NOTE: This test presently fails because apparantly,
    # SciHub's `intersects` parameter does not work reliably.
    #
    def test_spatial_queries(self):
        loc, ref_coords = next(iter(config.CONFIG['LOCATIONS'].items()))
        with self.subTest(location=loc):
            file_list = scihub.search(
                {'location': [loc], 'to_time': '2017-09-01T00:00:00Z'},
                server='S3', limit=20)
            for f in file_list:
                with self.subTest(product=f['filename']):
                    #
                    # Assert that the products indeed intersect the
                    # requested location.
                    #
                    self.assertTrue(geo.intersect(f['coords'], ref_coords,
                                                  tolerance=0.1))

    def test_get_file_list(self):
        q = {'mission': 'Sentinel-3'}
        limit = 107
        file_list = scihub.search(q, limit=limit)
        #
        # Assert that only `limit` entries are returned.
        #
        self.assertEqual(limit, len(file_list))
        for f in file_list:
            #
            # Assert that each entry contains the attributes `url`, `uuid` and
            # `filename`.
            #
            self.assertIn('url', f)
            self.assertIn('uuid', f)
            self.assertIn('filename', f)


# -----------------------------------------------------------------------------
# SCIHUB DOWNLOAD
# -----------------------------------------------------------------------------
class ScihubDownloadTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.clear_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_download(self):
        file_list = scihub.search({'query': SMALL_SIZE_QUERY}, limit=1)
        for f in file_list:
            with self.subTest(url=f['url']):
                result = scihub.download(f)
                #
                # Assert that the download didn't fail and that
                # the returned file path exists.
                #
                self.assertNotEqual(result, False)
                self.assertTrue(os.path.isfile(result))

    def test_download_many(self):
        file_list = scihub.search({'query': SMALL_SIZE_QUERY},
                                  limit=2)
        scihub.download_many(file_list)
        #
        # Assert that all downloads were successful.
        #
        local_files = helpers.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        local_files_identifiers = [os.path.splitext(os.path.split(_)[1])[0]
                                   for _ in local_files]
        for f in file_list:
            self.assertIn(f['filename'], local_files_identifiers)
        for f in local_files:
            with self.subTest(file=f):
                _, healthy, msg = check.check_file(f, mode='file')
                helpers.eprint(msg)
                self.assertTrue(healthy)

    def test_redownload(self):
        test_config.copy_corrupt_data()
        local_files = helpers.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        scihub.redownload(local_files)
        new_local_files = helpers.ls(config.CONFIG['GENERAL']['DATA_DIR'])
        self.assertEqual(set(local_files), set(new_local_files))
        for f in local_files:
            with self.subTest(file=f):
                _, healthy, msg = check.check_file(f, mode='file')
                helpers.eprint(msg)
                self.assertTrue(healthy)


# -----------------------------------------------------------------------------
# CHECK
# -----------------------------------------------------------------------------
class CheckTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_check_file_md5_healthy(self):
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files check out in `md5` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='md5')
                    self.assertTrue(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_zip_healthy(self):
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files check out in `file` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='file')
                    self.assertTrue(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_md5_corrupt(self):
        test_config.clear_test_data()
        test_config.copy_corrupt_data()
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files are detected as corrupt in `md5` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='md5')
                    self.assertFalse(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))

    def test_check_file_zip_corrupt(self):
        test_config.clear_test_data()
        test_config.copy_corrupt_data()
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the files are detected as corrupt in `file` mode.
                #
                try:
                    file_path, healthy, message = \
                        check.check_file(f, mode='file')
                    self.assertFalse(healthy)
                except Exception as e:
                    self.fail('File check failed: {}'.format(e))


# -----------------------------------------------------------------------------
# CHECKSUM
# -----------------------------------------------------------------------------
class ChecksumTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_test_data()

    def test_md5(self):
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the md5 checksum returned by checksum.md5() is
                # equal to the md5 sum returned by bash md5 or md5sum tool.
                #
                for exe in ['md5', 'md5sum']:
                    if helpers._which(exe) is not None:
                        bash_output = subprocess.check_output([exe, f])
                        if not PY2:
                            bash_output = bash_output.decode()
                        bash_md5 = re.search('[a-zA-Z0-9]{32}',
                                             bash_output).group()
                        break
                self.assertEqual(
                    checksum.md5(f), bash_md5
                )

    def test_etag_small_files(self):
        for f in helpers.ls(config.CONFIG['GENERAL']['DATA_DIR']):
            with self.subTest(file=f):
                #
                # Assert that the computed etag is equal to the md5
                # checksum for files smaller than the chunksize.
                #
                size_mb = max(10, int(os.path.getsize(f) / 1024**2))
                self.assertEqual(
                    checksum.md5(f), checksum.etag(f, chunksize=2 * size_mb)
                )

    # def test_etag_large_files(self):
    #     pass


# -----------------------------------------------------------------------------
# MAIN
# -----------------------------------------------------------------------------
class MainTestCase(TestCase):

    @classmethod
    def setUpClass(cls):
        test_config.set_test_config()
        cls.check_mode = config.CONFIG['GENERAL']['CHECK_MODE']
        config.CONFIG['GENERAL']['CHECK_MODE'] = 'file'

    @classmethod
    def tearDownClass(cls):
        config.CONFIG['GENERAL']['CHECK_MODE'] = cls.check_mode

    def setUp(self):
        test_config.copy_test_data()

    def tearDown(self):
        test_config.clear_all()

    def test_ls(self):
        q = {'time': 'today', 'satellite': 'S3A',
             'location': ['Ireland_Mace_Head']}
        files = scihub.search(q)
        result = main.ls(q)
        self.assertEqual(len(result), len(files))

    def test_get(self):
        test_config.clear_test_data()
        q = {'satellite': 'S3A', 'query': SMALL_SIZE_QUERY}
        files = scihub.search(q, limit=2)
        main.get(q, limit=2)

        for f in files:
            ext = '.zip'
            with self.subTest(product=f['filename']):
                self.assertTrue(
                    os.path.isfile(os.path.join(
                        config.CONFIG['GENERAL']['DATA_DIR'],
                        f['filename']) + ext)
                )

    def test_doctor(self):
        test_config.copy_corrupt_data()
        corrupt_files = helpers.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                   path=False)
        healthy_files = helpers.ls(test_config.TEST_DATA_DIR_ORIGINAL,
                                   path=False)
        main.doctor()
        #
        # Assert that the number of healthy/corrupt files detected are correct
        #
        self.assertEqual(check.BAD_FILE_COUNTER, len(corrupt_files))
        self.assertEqual(check.FILE_COUNTER,
                         len(corrupt_files) + len(healthy_files))
        bad_files = [os.path.split(_)[1] for _ in check.BAD_FILES]
        for corrupt_file in corrupt_files:
            #
            # Assert that each corrupt file has been registered.
            #
            self.assertIn(corrupt_file, bad_files)

    def test_doctor_delete(self):
        test_config.copy_corrupt_data()
        corrupt_files = helpers.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                   path=False)
        healthy_files = helpers.ls(test_config.TEST_DATA_DIR_ORIGINAL,
                                   path=False)
        main.doctor(delete=True)
        #
        # Assert that the corrupt files have been deleted.
        #
        for f in corrupt_files:
            self.assertFalse(os.path.isfile(os.path.join(
                    config.CONFIG['GENERAL']['DATA_DIR'], f)))

        #
        # Assert that the healthy files have not been deleted.
        #
        for f in healthy_files:
            self.assertTrue(os.path.isfile(os.path.join(
                    config.CONFIG['GENERAL']['DATA_DIR'], f)))

    def test_doctor_repair(self):
        test_config.copy_corrupt_data()
        corrupt_files = helpers.ls(test_config.TEST_DATA_DIR_CORRUPT,
                                   path=False)
        # healthy_files = helpers.ls(test_config.TEST_DATA_DIR_ORIGINAL,
        #                            path=False)
        main.doctor(repair=True)

        for f in corrupt_files:
            repaired_f = os.path.join(config.CONFIG['GENERAL']['DATA_DIR'], f)
            with self.subTest(file=repaired_f):
                #
                # Assert that each corrupt file has been repaired.
                #
                _, healthy, msg = check.check_file(repaired_f, mode='file')
                helpers.eprint(msg)
                self.assertTrue(healthy)


# -----------------------------------------------------------------------------
# HELPERS
# -----------------------------------------------------------------------------

# class HelpersTestCase(TestCase):
#
#     def setUp(self):
#         test_config.set_test_config()
#         test_config.copy_test_data()
#
#     def tearDown(self):
#         test_config.clear_all()



# coding=utf-8
""" This module implements the interface with the SciHub portal,
    especially authentication and queries.

TODO: Switch from multiprocessing to threading?

"""
import lxml.etree as ET
import os
import sys
import time
import logging
import re
import errno
# import hashlib
import base64
import ssl
# import math
import datetime as DT
import pytz
import multiprocessing
import socket
from functools import partial
from collections import OrderedDict
from .config import CONFIG
from . import helpers, tty, geo, checksum

logger = logging.getLogger('esahub')
PY2 = sys.version_info < (3, 0)

if PY2:
    # Python2 imports
    from urllib import urlencode
    from urllib2 import urlopen, Request
    from urllib2 import HTTPError, URLError
    from urlparse import urlparse, parse_qs
else:
    # Python3 imports
    from urllib.request import urlopen, Request
    from urllib.parse import urlparse, parse_qs, urlencode
    from urllib.error import HTTPError, URLError

COUNTER = 0
TOTAL = 0
COLLECT = []

DOWNLOAD_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/$value"
CHECKSUM_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/Checksum/Value/$value"
PREVIEW_URL_PATTERN = \
    "{host}/odata/v1/Products('{uuid}')/Products('Quicklook')/$value"


# -----------------------------------------------------------------------------
# CLASS DEFINITION: CONNECTION
# -----------------------------------------------------------------------------
class NotFoundError(Exception):
    pass


class Connection(object):
    """
    An abstraction of an authenticated connection to the Copernicus SciHub.
    """
    user = ''
    password = ''

    def __init__(self, mission=None, user=None, password=None, host=None):
        if mission is not None:
            self.authenticate(mission=mission)
        else:
            self.login(user=user, password=password, host=host)

    # @classmethod
    def login(self, user, password, host):
        """Define the login credentials.

        Parameters
        ----------
        user : str
        password : str
        host : str
            The URL to the SciHub host.
        """
        self.host = host
        self.user = user
        self.password = password

    # @classmethod
    def authenticate(self, mission):
        """Sets the login credentials as appropriate for the specified mission.

        Parameters
        ----------
        mission : {'S1A', 'S1B', 'S2A', 'S2B', 'S3A'}
        """
        if mission in CONFIG['SATELLITES']:
            source = CONFIG['SATELLITES'][mission]['source'][0]
            self.login(
                **CONFIG['SERVERS'][source]
            )
        else:
            logging.warning('No server found for mission: {}'.format(mission))

    # GENERAL METHODS FOR HTTP REQUESTS AND DOWNLOADS
    # -------------------------------------------------------------------------
    @classmethod
    def _encode(cls, url):
        """Minimal URL encoding as needed for this specific purpose.

        Parameters
        ----------
        url : str

        Returns
        -------
        str
            The encoded URL.
        """
        url = url.replace(' ', '%20')
        url = url.replace('"', '%22')
        return url

    # @classmethod
    def resolve(self, url):
        """Resolves the specified URL handling the HTTP authentication.

        Authenticates with the server and returns the urllib response object.
        Handles connection timeouts and 503 status codes as specified in the
        config module.

        Parameters
        ----------
        url : str

        Returns
        -------
        HTTP response
            Provides functions such as `read()`, `readline()`, `getcode()`,
            `geturl()`, `info()`

        """
        trials = 0
        while True:
            try:
                url = self._encode(url)
                request = Request(url)
                if not (self.user is None and self.password is None):
                    if PY2:
                        base64string = base64.b64encode('{}:{}'.format(
                                self.user, self.password))
                    else:
                        base64string = base64.b64encode('{}:{}'.format(
                                self.user, self.password
                            ).encode('ascii')).decode('ascii')
                    request.add_header("Authorization",
                                       "Basic {}".format(base64string))
                gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
                result = urlopen(request, timeout=CONFIG['GENERAL']['TIMEOUT'],
                                 context=gcontext)
            except socket.timeout as e:
                return HTTPError(url=url, code=504, msg='{}'.format(e),
                                 hdrs={}, fp='')
            except HTTPError as e:
                # print(e)
                # if CONFIG['GENERAL']['WAIT_ON_503'] and e.code == 503:
                #     time.sleep(CONFIG['GENERAL']['RECONNECT_TIME'])
                if CONFIG['GENERAL']['WAIT_ON_503']:
                    time.sleep(CONFIG['GENERAL']['RECONNECT_TIME'])
                else:
                    trials += 1
                    #
                    # If the connection could not be established after the
                    # maximum number of trials, raise with error message.
                    #
                    if trials >= CONFIG['GENERAL']['TRIALS']:
                        logging.error('{0} {1}, URL: {2}'.format(
                            tty.error('ERROR (%i trials):' % trials), e, url))
                        # raise
                        return e
                    #
                    # Wait for an increasing time after every trial.
                    #
                    time.sleep(CONFIG['GENERAL']['RECONNECT_TIME'] * trials)
            else:
                return result

    def auto_resolve(self, url):
        # Detect server from url:
        # host = re.search('(.*)/search', url).group(1)
        # auth = helpers.select(CONFIG['SERVERS'], first=True, host=host)
        auth = {}
        for servername, server in CONFIG['SERVERS'].items():
            if server['host'] in url:
                auth = server
                break
        # if len(auth) == 0:
        #     print(url)
        self.login(**auth)
        return self.resolve(url)


# -----------------------------------------------------------------------------
# NEW METHODS
# -----------------------------------------------------------------------------
def _generate_next_url(url, total=None):
    parsed_url = urlparse(url)
    q_params = parse_qs(parsed_url.query)
    q_params = {k: v[0] if len(v) == 1 else v for k, v in q_params.items()}
    if 'rows' in q_params:
        q_params['rows'] = int(q_params['rows'])
    else:
        q_params['rows'] = 10
    if 'start' in q_params:
        q_params['start'] = int(q_params['start'])
    else:
        q_params['start'] = 0
    q_params['start'] += q_params['rows']

    if total is None or q_params['start'] <= total:
        parsed_url = parsed_url._replace(query=urlencode(q_params, doseq=True))
        return parsed_url.geturl()
    else:
        return False


def _parse_page(url, first=False):
    file_list = []

    prefixes = {
        'os': 'http://a9.com/-/spec/opensearch/1.1/',
        'opensearch': 'http://a9.com/-/spec/opensearch/1.1/',
        'doc': 'http://www.w3.org/2005/Atom',
        'gml': 'http://www.opengis.net/gml'
    }

    #
    # Retrials are handled in the resolve method
    #
    conn = Connection()
    try:
        response = conn.auto_resolve(url)
        if isinstance(response, HTTPError):
            raise response
    except HTTPError:
        if first:
            return 0, []
        else:
            return []

    xml = response.read()
    try:
        root = ET.fromstring(xml)
        total_results = int(root.find('os:totalResults', prefixes).text)
        # next_link = root.find("./doc:link[@rel='next']", prefixes)
        for entry in root.findall('doc:entry', prefixes):

            filename = entry.find("./doc:str[@name='identifier']",
                                  prefixes).text
            ingestiondate = helpers.to_date(
                entry.find("./doc:date[@name='ingestiondate']", prefixes).text,
                output='date')
            try:
                footprint_tag = entry.find("./doc:str[@name='gmlfootprint']",
                                           prefixes).text
                match = re.search('<gml:coordinates>(.*)</gml:coordinates>',
                                  footprint_tag)
                coords = geo.gml_to_polygon(match.groups()[0])
            except AttributeError:
                coords = None

            filesize = helpers.h2b(entry.find("./doc:str[@name='size']",
                                              prefixes).text)
            preview_url = entry.find("./doc:link[@rel='icon']",
                                     prefixes).attrib['href']
            file_dict = {
                # 'position'  : file_counter,
                'title': entry.find('doc:title', prefixes).text,
                'url': entry.find('doc:link', prefixes).attrib['href'],
                'host': conn.host,
                'preview': preview_url,
                'uuid': entry.find('doc:id', prefixes).text,
                'filename': filename,
                'size': filesize,
                'ingestiondate': ingestiondate,
                'coords': coords
            }
            file_list.append(file_dict)

    except ET.XMLSyntaxError:
        # not valid XML
        total_results = 0
        file_list = []

    if first:
        return total_results, file_list
    else:
        return file_list


def _get_file_list_from_url(url, limit=None):
    global COUNTER, TOTAL, COLLECT
    """ This function returns a list of all files resulting from the specified
    query.

    This method executes the query defined by the passed URL. It automatically
    takes care of the pagination implemented by SciHub and iterates over the
    returned pages.

    Parameters
    ----------
    query_url : str
        The SciHub URL that constitutes the query. It can be built using
        `Query.build_query_url()`
    limit : int, optional
        The maximum number of results to return. If `None`, return all results
        (default: None).

    Returns
    -------
    list of dict
        A list of dictionaries containing metadata for each found file. Each
        dictionary contains the following keys:
        `position`, `title`, `url`, `preview`, `id`, `filename`, `size`,
        `coords`
    """
    # file_list = []
    # file_counter = 0
    total_results = 0
    # TOTAL_SIZE = 0.0

    logging.debug('QUERYING {0} ...'.format(url))
    urlname = urlparse(url).netloc

    #
    # Parallel parsing of pages.
    #

    # Parse first page serially to get total number of results.
    current_url = url
    total_results, files_first_page = _parse_page_wrapper(current_url,
                                                          first=True)
    if limit is None:
        TOTAL = total_results
    else:
        TOTAL = min(limit, total_results)

    COUNTER = 0
    COLLECT = []

    def _callback(result):
        global COUNTER, TOTAL, COLLECT
        COLLECT.extend(result)
        COUNTER += len(result)
        status = (0, 1) if TOTAL == 0 else (COUNTER, TOTAL)
        tty.status('Querying {}...'.format(urlname), progress=status)

    _callback(files_first_page)

    #
    # The processes are not CPU-intensive at all (the bottleneck is the server
    # response time), therefore we can handle very large number of processes.
    #
    pool = multiprocessing.Pool(
            processes=CONFIG['GENERAL']['N_SCIHUB_QUERIES'])

    while True:
        current_url = _generate_next_url(current_url, total=TOTAL)
        if not current_url:
            break
        pool.apply_async(_parse_page_wrapper, (current_url,),
                         callback=_callback)

    pool.close()
    pool.join()
    tty.finish_status()

    if limit is not None:
        COLLECT = COLLECT[:limit]

    # from IPython import embed; embed()

    # TOTAL_SIZE = sum([f['size'] for f in COLLECT])

    return COLLECT


def _parse_page_wrapper(*args, **kwargs):
    #
    # For some weird reason, _parse_page sometimes raises:
    #      AttributeError: 'str' object has no attribute 'close'
    # from the tempfile module. As this only occurs sometimes,
    # just try again ...
    #
    while True:
        try:
            result = _parse_page(*args, **kwargs)
        except Exception as e:
            time.sleep(1)
        else:
            return result


def _build_query(query={}):
    """ Builds and returns the query URL given the command line input
    parameters.

    Parameters
    ----------
    query : list, optional
        A list of additional custom query elements submitted to SciHub. The
        queries will be concatenated with ampersands (&).
    """
    query_list = []
    sort_string = ''

    # Default ingestiontime query parameters
    start = '1970-01-01T00:00:00.000Z'
    end = 'NOW'
    # Geospatial query list
    geo_query_list = []

    for key, val in query.items():
        if key == 'mission':
            # Build mission selection query:
            query_list.append('platformname:{0}'.format(query['mission']))

        elif key == 'satellite':
            # Build satellite selection query:
            query_list.append('identifier:{0}*'.format(query['satellite']))

        elif key == 'from_time':
            start = query['from_time']
        elif key == 'to_time':
            end = query['to_time']

        elif key == 'time':
            if val == 'today':
                start = DT.datetime.strftime(DT.datetime.now(pytz.utc),
                                             '%Y-%m-%dT00:00:00.000Z')
            elif val == 'yesterday':
                start = DT.datetime.strftime(DT.datetime.now(pytz.utc)
                                             - DT.timedelta(1),
                                             '%Y-%m-%dT00:00:00.000Z')
                end = DT.datetime.strftime(DT.datetime.now(pytz.utc),
                                           '%Y-%m-%dT00:00:00.000Z')
            elif val == 'midnight':
                end = DT.datetime.strftime(DT.datetime.now(pytz.utc),
                                           '%Y-%m-%dT00:00:00.000Z')
            elif val == '24h':
                start = DT.datetime.strftime(DT.datetime.now(pytz.utc)
                                             - DT.timedelta(1),
                                             '%Y-%m-%dT%H:%M:%S.000Z')
                start = DT.datetime.strftime(DT.datetime.now(pytz.utc),
                                             '%Y-%m-%dT%H:%M:%S.000Z')

        elif key == 'geo':
            # Build geospatial query:
            if type(val) is not list:
                val = [val]
            for item in val:
                geo_query_list.append(
                        'footprint:"Intersects({0})"'.format(item))

        elif key == 'location':
            if type(val) is not list:
                val = [val]
            for loc in val:
                if loc in CONFIG['LOCATIONS']:
                    geo_query_list.append('footprint:"Intersects({0})"'.format(
                            CONFIG['LOCATIONS'][loc]))
                else:
                    logging.error(
                        '{0} {1}'.format(tty.error('Location not found:'), loc)
                    )

        elif key == 'type':
            query_list.append('producttype:{0}'.format(query['type']))

        elif key == 'query':
            query_list.append('{0}'.format(query['query']))

        elif key == 'sort':
            sort_string = '&orderby={} {}'.format(*query['sort'])

        # Not a special keyword. Pass directly to SciHub
        else:
            query_list.append('{}:{}'.format(key, val))

    # Build ingestiondate query:
    query_list.append('ingestiondate:[{0} TO {1}]'.format(start, end))

    # Build geospatial query:
    if len(geo_query_list):
        query_list.append(
            '({})'.format(' OR '.join(geo_query_list))
        )

    # Generate full query string
    if len(query_list) == 0:
        query_list.append('*:*')
    query_string = ' AND '.join(query_list)

    #
    # Example: 'https://scihub.copernicus.eu/s3/search?
    #           q=(footprint:"Intersects(POLYGON((-25.100 46.800,
    #           -5.250 46.800,-5.250 57.400,-25.100 57.400,-25.100 46.800)))")
    #           &rows=25&start=0'
    #
    query_url = 'q={q}&start=0&rows={rows}{sort}'.format(
        q=query_string,
        rows=CONFIG['GENERAL']['ENTRIES'],
        sort=sort_string
    )
    return query_url


def _build_url(query, server):
    if type(query) is dict:
        query_string = _build_query(query)
    else:
        query_string = query
    return '{url}/search?{q}'.format(
        url=CONFIG['SERVERS'][server]['host'], q=query_string
    )


def _download(url, destination, quiet=None, queue=None):
    """Downloads a file from the remote server into the specified destination.

    Parameters
    ----------
    url : str
        The source URL.
    destination : str
        The local target file path.
    quiet : bool, optional
        Whether to suppress any command line output.
    queue : multiprocessing.Manager.Queue, optional
        A multiprocessing queue to submit status messages.

    Returns
    -------
    bool
        True if successful, False otherwise.
    """
    quiet = CONFIG['GENERAL']['QUIET'] if quiet is None else quiet

    #
    # Create directory.
    #
    path, file_name = os.path.split(destination)
    try:
        os.makedirs(path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise

    try:
        response = resolve(url)
        if PY2:
            size = int(response.info().getheaders("Content-Length")[0])
        else:
            size = int(response.headers.get("Content-Length"))
        # size_str = helpers.b2h(size)
        if queue is not None:
            queue.put((file_name, {'total': size,
                                   'desc': 'Downloading {name}'}))
        CHUNK = 32 * 1024
        chunks_done = 0
        t_start_download = time.time()
        with open(destination, 'wb') as f:
            while True:
                chunk = response.read(CHUNK)
                if not chunk:
                    break
                f.write(chunk)
                chunks_done += 1
                if queue is not None:
                    queue.put((file_name, {'progress': CHUNK}))
                    # if size is not None:
                    #     queue.put((file_name,
                    #                '{} Downloading ({} / {})'.format(
                    #                     file_name,
                    #                     helpers.b2h(chunks_done*CHUNK),
                    #                     size_str)
                    #                ))
                    # else:
                    #     queue.put((file_name, '{} Downloading ({})'.format(
                    #             file_name, helpers.b2h(chunks_done*CHUNK))))
        t_stop_download = time.time()
        download_size = os.path.getsize(destination)
        download_rate = download_size/(t_stop_download - t_start_download)
        msg = '{0} Downloaded ({1} at {2}/s)'.format(
            file_name, helpers.b2h(download_size), helpers.b2h(download_rate)
        )
        if not quiet:
            logging.info(msg)
        if queue is not None:
            queue.put((file_name, {'desc': '{name} Downloaded'}))
        #     queue.put((file_name, msg))

    except HTTPError as e:
        return False
    except URLError as e:
        return False
    except Exception as e:
        logging.error(repr(e))
        return False
    else:
        return True


# def _get_file_list_wrapper(url):
#     host = re.search('(.*)/search',url).group(1)
#     server = helpers.select(list(CONFIG['SERVERS'].values()), host=host)
#     if type(server) is list:
#         server = server[0]
#     Connection.login(**server)
#     return _get_file_list_from_url(url)


def _get_available_servers():
    servers = [server for server, conf in CONFIG['SERVERS'].items()
               if len(conf['user']) > 0 and len(conf['password']) > 0]
    return servers


def _ping_single(servername):
    # print('Pinging {} ...'.format(servername))
    server = CONFIG['SERVERS'][servername]
    conn = Connection(**server)
    try:
        result = conn.resolve('{url}/search?q=*:*'.format(url=server['host']))
        # results[servername] = result.status
        return (servername, result.status)
    except HTTPError as e:
        # results[servername] = e.status
        return (servername, e.status)


def _auto_detect_server_from_query(query, available_only=False):
    servers = None

    if 'satellite' in query and query['satellite'] in CONFIG['SATELLITES']:
        servers = CONFIG['SATELLITES'][query['satellite']]['source']

    if 'identifier' in query:
        sat = query['identifier'][:3]
        if sat in CONFIG['SATELLITES']:
            servers = CONFIG['SATELLITES'][sat]['source']

    if 'mission' in query:
        sats = helpers.select(CONFIG['SATELLITES'], platform=query['mission'])
        if len(sats) > 0:
            ll = [source for sat in sats.values() for source in sat['source']]
            servers = list(OrderedDict.fromkeys(ll))

    #
    # If the server couldn't be determined from the query, return a list of
    # all servers.
    #
    if servers is None:
        servers = list(CONFIG['SERVERS'].keys())

    if available_only:
        servers = [server for server in servers
                   if server in _get_available_servers()]

    return servers


def _uuid_from_identifier(identifier):
    identifier = os.path.splitext(os.path.split(identifier)[1])[0]
    results = search({'identifier': identifier+'*'})
    if len(results) == 0:
        raise NotFoundError('Product not found: {}'.format(identifier))
    return results[0]['uuid']


def _host_and_uuid_from_identifier(identifier):
    identifier = os.path.splitext(os.path.split(identifier)[1])[0]
    results = search({'identifier': identifier+'*'})
    if len(results) == 0:
        raise NotFoundError('Product not found: {}'.format(identifier))
    return (results[0]['host'], results[0]['uuid'])


def _host_from_uuid(uuid):
    #
    # IMPLEMENT THIS
    #
    return None


def _download_url_from_identifier(identifier):
    host, uuid = _host_and_uuid_from_identifier(identifier)
    return _download_url_from_uuid(uuid, host=host)


def _checksum_url_from_identifier(identifier):
    host, uuid = _host_and_uuid_from_identifier(identifier)
    return _checksum_url_from_uuid(uuid, host=host)


def _preview_url_from_identifier(identifier):
    host, uuid = _host_and_uuid_from_identifier(identifier)
    return _preview_url_from_uuid(uuid, host=host)


def _download_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return DOWNLOAD_URL_PATTERN.format(host=host, uuid=uuid)


def _checksum_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return CHECKSUM_URL_PATTERN.format(host=host, uuid=uuid)


def _preview_url_from_uuid(uuid, host=None):
    if host is None:
        host = _host_from_uuid(uuid)
    return PREVIEW_URL_PATTERN.format(host=host, uuid=uuid)


# -----------------------------------------------------------------------------
# API METHODS
# -----------------------------------------------------------------------------
def resolve(url, server=None):
    """ Resolves a SciHub URL and automatically determines the correct HTTP
    authentication.

    Parameters
    ----------
    url : str
        The URL to resolve.
    server : str, optional
        If specified, resolve using the specified server. Not in general
        needed.

    Returns
    -------
    HTTPResponse
        The HTTPRessponse or HTTPError object resulting from the URL.
    """
    if server is None:
        if 'server' in CONFIG['GENERAL']['QUERY']:
            auth = CONFIG['SERVERS'][CONFIG['GENERAL']['QUERY']['server']]
            conn = Connection(**auth)
            return conn.resolve(url)
        else:
            conn = Connection()
            return conn.auto_resolve(url)
    else:
        auth = CONFIG['SERVERS'][server]
        conn = Connection(**auth)
        return conn.resolve(url)


def ping(server=None):
    results = {}
    if server is not None:
        servers = [server]
    else:
        servers = list(CONFIG['SERVERS'].keys())

    pool = multiprocessing.Pool(processes=len(servers))
    results = pool.map_async(_ping_single, servers)
    # pool.close()
    # pool.join()
    # for servername,server in servers.items():
    while not results.ready():
        time.sleep(1)
    return results.get()


def search(query={}, server='auto', limit=None, **kwargs):
    """ Search SciHub for satellite products.
    Parameters
    ----------
    query : dict, optional
        Query parameters as dictionary. May contain key 'server' (see below).
    server : str, optional
        If server is 'all', search all saved SciHub servers.
        If server is 'auto', attempt to auto-detect the server from the query.
        Otherwise, use the specified server (as configured in CONFIG.SERVERS)
        (default: 'auto')
    limit : int, optional
        The maximum number of results to return.
    kwargs : dict, optional
        The query parameters can also be passed as keyword arguments.

    Returns
    -------
    list of dict
        A list of dictionaries representing the found search results.
    """
    #
    # Search in each server.
    #
    query.update(kwargs)
    if 'server' in query:
        server = query.pop('server')

    if server == 'all':
        servers = _get_available_servers()
    elif server == 'auto':
        servers = _auto_detect_server_from_query(query, available_only=True)
    else:
        servers = [server]

    if servers is None:
        servers = []
    results = []
    query_string = _build_query(query)

    remaining = limit
    for servername in servers:
        server = CONFIG['SERVERS'][servername]
        # Connection.login(**server)
        url = '{url}/search?{q}'.format(
            url=server['host'], q=query_string
        )
        logging.debug('Trying server {}: {}'.format(servername, url))
        results.extend(_get_file_list_from_url(url, limit=remaining))
        if remaining is not None:
            remaining = limit - len(results)
            if remaining <= 0:
                break

    #
    # Delete duplicate results (if product is on multiple servers).
    #
    filtered = OrderedDict(zip([_['filename'] for _ in results], results))
    return list(filtered.values())


# def remote_md5(query_url):
#     """Returns the md5 sum of the file stored on SciHub given the download
#     URL.
#
#     Parameters
#     ----------
#     query_url : str
#         The SciHub download URL.
#
#     Returns
#     -------
#     str
#         The md5 checksum in lower case.
#     """
#
#     md5_query_url = query_url.replace('/$value','/Checksum/Value/$value')
#     response = Connection.resolve(md5_query_url)
#     if PY2:
#         return response.read().lower()
#     else:
#         return response.read().decode().lower()


def md5(product=None, uuid=None):
    """Returns the md5 sum of the file stored on SciHub given the product name
    or uuid.

    Parameters
    ----------
    product : str, optional
        The product name. If given, `uuid` is ignored.
    uuid : str, optional
        The product uuid.

    Returns
    -------
    str
        The md5 checksum in lower case.
    """
    if product is not None:
        if type(product) is dict and 'uuid' in product and 'host' in product:
            md5_url = _checksum_url_from_uuid(product['uuid'],
                                              host=product['host'])
        elif type(product) is str:
            md5_url = _checksum_url_from_identifier(product)
        else:
            return False

    elif uuid is not None:
        md5_url = _checksum_url_from_uuid(uuid)

    response = resolve(md5_url)
    if PY2:
        return response.read().lower()
    else:
        return response.read().decode().lower()


def exists(product):
    search_results = search({'identifier': product+'*'}, limit=1)
    return len(search_results) > 0


def download(product, queue=None, stopqueue=True):
    """Download a satellite product.

    Checks for file existence and MD5 checksum.

    Parameters
    ----------
    product : str or dict
        The name of the product to be downloaded from SciHub.
        Alternatively, a dictionary representing a search result from SciHub.
    queue : multiprocessing.Manager.Queue, optional
        A multiprocessing queue to submit status messages.
    stopqueue : bool, optional
        Whether to send a poison pill upon completion.

    Returns
    -------
    str
        The local file path if the download was successful OR the file already
        exists and passes the md5 checksum test. False otherwise.
    """
    if type(product) is dict:
        fdata = product
    else:
        fdata = search({'identifier': os.path.splitext(product)[0]+'*'})[0]

    satellite = helpers.get_satellite(fdata['filename'])
    ext = CONFIG['SATELLITES'][satellite]['ext']
    # suffix = CONFIG['SATELLITES'][satellite]['suffix']
    file_name = fdata['filename'] + ext
    # file_name = os.path.splitext(product)[0] + '.zip'

    b_download = True
    b_file_okay = False
    complete = False

    full_file_path = os.path.join(CONFIG['GENERAL']['DATA_DIR'], file_name)

    #
    # Establish authentication depending on the mission (S1/S2/S3)
    #
    # Connection.authenticate(satellite)

    #
    # Check if file already exists in location:
    #
    if os.path.exists(full_file_path) and os.path.isfile(full_file_path):

        if not CONFIG['GENERAL']['CHECK_EXISTING']:
            msg = '{} Skipping existing - MD5 not checked'.format(file_name)
            logging.debug(msg)
            if queue is not None:
                queue.put((file_name, {'desc': 'Skipping {name}'}))
            #     queue.put((file_name, msg))
            b_download = False
        else:
            if queue is not None:
                queue.put((file_name,
                          {'desc': 'Checking {name} ...'}))
            #     queue.put((file_name,
            #                '{} Exists - checking MD5 ...'.format(file_name)))
            local_md5 = checksum.md5(full_file_path)
            remote_md5 = md5(fdata)
            if local_md5 == remote_md5:
                msg = '{} Skipping download (MD5 okay)'.format(file_name)
                logging.debug(msg)
                if queue is not None:
                    queue.put((file_name, {'desc': 'Skipping {name}'}))
                #     queue.put((file_name, msg))
                b_download = False
                b_file_okay = True
            else:
                msg = '{} MD5 wrong: redownloading ...'.format(file_name)
                logging.debug(msg)
                # if queue is not None:
                #     queue.put((file_name, msg))

    # if b_download and CONFIG['GENERAL']['SKIP_EXTRACTED'] and \
    #         index.exists(file_name, where='local'):
    #     msg = '{} Skipping download (index exists)'.format(file_name)
    #     logging.debug(msg)
    #     if queue is not None:
    #         queue.put((file_name, msg))
    #     b_download = False

    # if b_download and CONFIG['GENERAL']['SKIP_UPLOADED'] and \
    #         storage.exists(file_name):
    #     msg = '{} Skipping download (already stored)'.format(file_name)
    #     logging.debug(msg)
    #     if queue is not None:
    #         queue.put((file_name, msg))
    #     b_download = False

    if b_download:
        # if queue is not None:
        #     queue.put((file_name, '{} Downloading ... '.format(file_name)))

        #
        # Retrials in case the MD5 hashsum fails
        #
        for i in range(CONFIG['GENERAL']['TRIALS']):
            try:
                complete = _download(fdata['url'], full_file_path, queue=queue)
            except Exception as e:
                logging.debug('{}: {}'.format(file_name, e))
                complete = False

            #
            # After download, check MD5 hashsum
            #
            if not complete:
                logging.debug('{} Download failed, trial {:d}/{:d}.'.format(
                        file_name, i+1, CONFIG['GENERAL']['TRIALS']))
                if queue is not None:
                    queue.put((file_name, {'desc': 'Retrying {name}'}))
                #     queue.put((file_name,
                #                '{} Retrying (trial {:d}/{:d})'
                #                .format(file_name, i+1,
                #                        CONFIG['GENERAL']['TRIALS'])))
            else:
                # if queue is not None:
                #     queue.put((file_name,
                #                '{} Downloaded - Checking MD5...'
                #                .format(file_name)))
                local_md5 = checksum.md5(full_file_path)
                remote_md5 = md5(fdata)
                if local_md5 != remote_md5:
                    logging.debug(
                        '{} MD5 checksum failed, trial {:d}/{:d}.'.format(
                            file_name, i+1, CONFIG['GENERAL']['TRIALS']))
                    if queue is not None:
                        queue.put((file_name,
                                  {'desc': 'Retrying {name}'}))
                    #     queue.put((file_name,
                    #                '{} Retrying ... (trial {:d}/{:d})'
                    #                .format(file_name, i+1,
                    #                        CONFIG['GENERAL']['TRIALS'])))
                else:
                    logging.debug('{} MD5 okay'.format(file_name))
                    if queue is not None:
                        queue.put((file_name, {'desc': '{name} okay'}))
                    #     queue.put((file_name,
                    #                '{} Downloaded - MD5 okay'
                    #                .format(file_name)))
                    b_file_okay = True
                    break

        if not b_file_okay:
            msg = '{} Download failed.'.format(file_name)
            logging.warning(msg)
            if queue is not None:
                queue.put((file_name, {'desc': '{name} Failed'}))
            #     queue.put((file_name, msg))

    if CONFIG['GENERAL']['DOWNLOAD_PREVIEW']:
        full_preview_path = os.path.join(CONFIG['GENERAL']['DATA_DIR'],
                                         fdata['filename']+'.jpeg')

        #
        # If not yet done, download preview file
        #
        if not os.path.exists(full_preview_path) or \
                not os.path.isfile(full_preview_path):
            if not download(fdata['preview'], full_preview_path, quiet=True):
                logging.info('  Preview not available.')

    #
    # Download done (or failed).
    # Send poison pill to parent process.
    #
    if stopqueue and queue is not None:
        queue.put((file_name, ))

    if b_file_okay:
        #
        # File has been downloaded successfully OR already exists
        # --> Return the file path
        #
        logging.debug('Download successful: {}'.format(full_file_path))
        return full_file_path

    elif b_download and not complete:
        #
        # File download failed --> Return FALSE
        #
        logging.error('Download failed: {}'.format(file_name))
        if queue is not None:
            queue.put((file_name, {'desc': '{name} Failed'}))
        #     queue.put((file_name, 'Download failed: {}'.format(file_name)))
        return False

    else:
        #
        # File download was skipped --> Return FALSE
        #
        return False


def download_many(products):
    """Downloads all requested files in parallel, checking for correct MD5 sums.

    Parameters
    ----------
    file_list : list of dict
        A list of dictionaries for each requested file. Can be obtained from
        `scihub_connect.Query.get_file_list()`
    """
    #
    # Parallel downloads
    #
    counter = 0
    total = len(products)
    q = multiprocessing.Manager().Queue()
    pool = multiprocessing.Pool(processes=CONFIG['GENERAL']['N_DOWNLOADS'])
    fn = partial(download, queue=q)
    result = pool.map_async(fn, products)
    while not result.ready():
        while True:
            try:
                msg = q.get(timeout=0.5)
                # fname, msg = q.get(timeout=0.5)
            except multiprocessing.queues.Empty:
                break
            else:
                # if msg is None:
                if len(msg) == 1:
                    counter += 1
                    tty.set_progress(counter, total)
                else:
                    fname, value = msg
                    tty.update(fname, value)
                    # tty.update(fname, msg)
    result.wait()
    tty.finish_status()
    tty.result('All done!')


def redownload(local_file_list):
    """Redownload a list of corrupt files.

    This method will automatically determine the correct source of each file
    from the file name. It will then attempt to redownload the files.

    Parameters
    ----------
    local_file_list : list of str
        A list of local files to be redownloaded from the server.
    """
    def _get_remote_files(local_files, satellite):
        product_names = [os.path.splitext(os.path.split(fpath)[1])[0]
                         for fpath in local_files]
        chunksize = 10
        remote_file_list = []
        for products in helpers.chunks(product_names, chunksize):
            query_string = '({})'.format(
                    ' OR '.join(['identifier:{}'.format(product)
                                 for product in products])
            )
            try:
                remote_file_list.extend(
                    search({'query': query_string, 'satellite': satellite})
                )
            except Exception as e:
                logging.warning('Skipping downloads: {}'.format(e))
        return remote_file_list

    remote_files = []
    for sat in CONFIG['SATELLITES']:
        files = [f for f in local_file_list
                 if os.path.split(f)[1].startswith(sat)]
        remote_files.extend(_get_remote_files(files, sat))

    logging.info('DOWNLOADING {}'.format(len(remote_files)))
    download_many(remote_files)



from .config import CONFIG
import sys
from tqdm import tqdm

BAR_FORMAT = "{desc} {percentage:3.0f}% |{bar}| " \
             "{n_fmt}/{total_fmt} " \
             "[{elapsed}, {rate_fmt: >8}{postfix}]"
# <{remaining}


def shorten(text, maxlen=30):
    if len(text) <= maxlen:
        return text
    else:
        n_end = int((maxlen - 3) / 2)
        n_start = maxlen - 3 - n_end
        return '{}...{}'.format(text[:n_start], text[-n_end:])


class Screen():
    _status = None
    _lines = {}

    def __init__(self, *args, **kwargs):
        self.reset_status(0)

    def reset_status(self, total):
        # status_fmt = ''
        if self._status is not None:
            self._status.close()
            self._status = tqdm(total=total,
                                disable=CONFIG['GENERAL']['QUIET'])

    def __setitem__(self, key, total):
        pass

    def update(self, key, msg):
        if key in self._lines:
            if 'progress' in msg:
                # Update progress bar
                self._lines[key].update(msg['progress'])
            if 'desc' in msg:
                desc = msg['desc'].format(name=shorten(key))
                self._lines[key].set_description_str(desc=desc)
            if 'total' in msg:
                self._lines[key].total = msg['total']
        else:
            # Create new progress bar
            if 'desc' in msg:
                desc = msg['desc'].format(name=shorten(key))
            else:
                desc = 'File {}'.format(len(self._lines) + 1)

            total = msg['total'] if 'total' in msg else None
            self._lines[key] = tqdm(
                desc=desc, total=total, unit='B', unit_scale=True,
                ascii=True, bar_format=BAR_FORMAT,
                mininterval=0.3,
                disable=CONFIG['GENERAL']['QUIET']
            )

    def set_status(self, counter, total=None, msg=None):
        if total is not None:
            self.reset_status(total)
        if msg is not None:
            self._status.desc = msg
        self._status.total

    def quit(self):
        # Close all progress bars
        if self._status is not None:
            del self._status
        for key, pbar in self._lines.items():
            del pbar

    def __del__(self):
        self.quit()


screen = None


def init():
    global screen
    screen = Screen()


def quit():
    global screen
    del screen


def wait_for_quit():
    global screen
    del screen


def update(key, msg):
    global screen
    screen.update(key, msg)


def finish_status():
    pass


def header(text):
    global screen
    screen['header'] = text


def result(text):
    global screen
    screen['result'] = text


def status(text, progress=None):
    global screen
    screen['status'] = text


def set_progress(counter, total):
    global screen


# TERMINAL OUTPUT FORMATTING
# -----------------------------------------------------------------------------
PURPLE = '\033[95m'
CYAN = '\033[96m'
DARKCYAN = '\033[36m'
BLUE = '\033[94m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
RED = '\033[91m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
END = '\033[0m'


def error(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, RED, message, END)
    else:
        return message


def warn(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, YELLOW, message, END)
    else:
        return message


def success(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, GREEN, message, END)
    else:
        return message


init()



from reprint import reprint
from reprint.reprint import is_atty, overflow_flag, last_output_lines
from reprint import output
import sys
import time


# -----------------------------------------------------------------------------
# Monkey patch reprint
# -----------------------------------------------------------------------------

def _new__setitem__(self, key, value):
    global is_atty
    with self.lock:
        super(output.SignalDict, self).__setitem__(key, value)
        if not is_atty:
            print("{}".format(value))
        else:
            self.parent.refresh(int(time.time()*1000), forced=False)


def _new_print_multi_line(content, force_single_line, sort_key):
    """
    'sort_key' 参数只在 dict 模式时有效
    'sort_key' parameter only available in 'dict' mode
    """

    global last_output_lines
    global overflow_flag
    global is_atty

    if not is_atty:
        if isinstance(content, list):
            for line in content:
                print(line)
        elif isinstance(content, dict):
            for k, v in sorted(content.items(), key=sort_key):
                print("{}".format(v))
        else:
            raise TypeError("Excepting types: list, dict. Got: {}".format(
                type(content)))
        return

    columns, rows = reprint.get_terminal_size()
    lines = reprint.lines_of_content(content, columns)
    if force_single_line is False and lines > rows:
        overflow_flag = True
    elif force_single_line is True and len(content) > rows:
        overflow_flag = True

    # 确保初始输出位置是位于最左处的
    # to make sure the cursor is at the left most
    print("\b" * columns, end="")

    if isinstance(content, list):
        for line in content:
            _line = reprint.preprocess(line)
            reprint.print_line(_line, columns, force_single_line)
    elif isinstance(content, dict):
        for k, v in content.items():
            _k, _v = map(reprint.preprocess, (k, v))
            reprint.print_line("{}".format(_v), columns, force_single_line)
    else:
        raise TypeError("Excepting types: list, dict. Got: {}".format(
            type(content)))

    # 输出额外的空行来清除上一次输出的剩余内容
    # do extra blank lines to wipe the remaining of last output
    print(" " * columns * (last_output_lines - lines), end="")

    # 回到初始输出位置
    # back to the origin pos
    print(reprint.magic_char * (max(last_output_lines, lines)-1), end="")
    sys.stdout.flush()
    last_output_lines = lines


output.SignalDict.__setitem__ = _new__setitem__
reprint.print_multi_line = _new_print_multi_line



# coding=utf-8
""" This module contains special functionality related to terminal output.
    It implements a simple command line GUI for the esahub package.
"""
from __future__ import print_function
from .config import CONFIG
# import struct
import sys
import time
import curses
import multiprocessing
from threading import Timer


def init():
    Screen.start()


def quit():
    Screen.quit()


def progressBar(progress, total, prefix='', suffix='', length=30, fill='#'):
    """Returns the string representation of a progress bar with `progress`
    filled.

    Parameters
    ----------
    progress : int
    total : int
    prefix : str, optional
    suffix : str, optional
    length : int, optional
        The length of the progress bar (default: 30)
    fill : char, optional
        The character constituting the progress bar (default: '#')

    Returns
    -------
    str
        A string representation of the progress bar.
    """
    if total == 0:
        fraction = 1.0
    else:
        fraction = min(1.0, progress/float(total))
    percent = ('{:5.0f}').format(100.0 * fraction)
    filledLength = int(length * fraction)
    bar = fill * filledLength + ' ' * (length - filledLength)
    return '{}[{}] {}%{}'.format(prefix, bar, percent, suffix)


class Screen:
    """A custom Screen class representing the GUI.
    """
    INITIATED = False
    LINE_COUNT = 0
    REFRESH_INTERVAL = 0.3
    CONTENT_START = 3
    CONTENT = []
    CONTENT_DICT = {}
    LAST_REFRESHED = time.time()
    WINDOW = None
    HEIGHT = 0
    WIDTH = 0
    LINES = 0
    TITLE = ''
    STATUS = ''
    PROGRESS = (0, 1)
    RESULT = ''
    PROMPT = ''
    SCROLL_POS = 0
    MAX_CONTENT_LENGTH = 0

    @classmethod
    def start(cls):
        """Start curses application."""
        cls.WINDOW = curses.initscr()
        cls.INITIATED = True
        cls.HEIGHT, cls.WIDTH = cls.WINDOW.getmaxyx()
        cls.LINES = cls.HEIGHT - cls.CONTENT_START - 4

        curses.noecho()
        curses.cbreak()
        curses.curs_set(0)
        cls.WINDOW.keypad(1)
        cls.WINDOW.idcok(False)
        cls.WINDOW.idlok(False)

        # Set color schemes
        # curses.start_color()
        # curses.init_pair(RED_ON_BLACK, curses.COLOR_RED, curses.COLOR_BLACK)

        cls.WINDOW.refresh()

        # REDRAW SCREEN EVERY `REFRESH_INTERVAL` SECONDS
        cls.time_refresh()

    @classmethod
    def time_refresh(cls):
        def wrapper():
            cls.time_refresh()
            cls.redraw()
        if cls.INITIATED:
            t = Timer(cls.REFRESH_INTERVAL, wrapper)
            t.start()
            return t
        else:
            return False

    @classmethod
    def paint(cls):
        """Paint the full GUI without explicity clearing the screen."""
        if cls.INITIATED:
            cls.HEIGHT, cls.WIDTH = cls.WINDOW.getmaxyx()
            cls.LINES = cls.HEIGHT - cls.CONTENT_START - 4

            cls.WINDOW.move(0, 0)
            # cls.WINDOW.refresh()

            cls.redraw_content()

            # Print title
            cls.set_header(cls.TITLE, force=True)
            # Print status
            cls.print_status_line()
            # Print result
            cls.set_result(cls.RESULT, force=True)
            # Print prompt
            cls.set_prompt(cls.PROMPT, force=True)

            cls.WINDOW.refresh()

    @classmethod
    def redraw_content(cls):
        """Draw the content lines from buffer."""
        if cls.INITIATED:

            # Print only lines from SCROLL_POS to SCROLL_POS + LINES
            counter = 0
            for pos, line in enumerate(
                cls.CONTENT[cls.SCROLL_POS:cls.SCROLL_POS+cls.LINES]
            ):
                cls.__print_line(cls.CONTENT_START+pos, line)
                counter += 1

            if counter < cls.LINES:
                for pos in range(counter, cls.LINES):
                    cls.__clear_line(cls.CONTENT_START+pos)

            cut_above = cls.SCROLL_POS
            cut_below = len(cls.CONTENT) - cls.LINES - cls.SCROLL_POS
            if cut_above > 0:
                cls.__print_line(2, '... {} more (press arrow up)'
                                    .format(cut_above))
            else:
                cls.__clear_line(2)
            if cut_below > 0:
                cls.__print_line(cls.HEIGHT-4, '... {} more (press arrow down)'
                                               .format(cut_below))
            else:
                cls.__clear_line(cls.HEIGHT-4)

            cls.WINDOW.refresh()

    @classmethod
    def redraw(cls):
        """Fully erase and redraw the GUI from the buffer."""
        if cls.INITIATED:
            lock = multiprocessing.Lock()
            lock.acquire()
            cls.WINDOW.erase()
            curses.doupdate()
            cls.paint()
            lock.release()

    @classmethod
    def update(cls, key, text):
        """Update the specified piece of content.

        Content is stored as a key-value pair, allowing to update the status
        for each line easily. Appends a new key if it does not yet exist.

        Parameters
        ----------
        key : str
        text : str
        """
        cls.MAX_CONTENT_LENGTH = max(cls.MAX_CONTENT_LENGTH, len(text))
        if key in cls.CONTENT_DICT:
            cls.set_line(cls.CONTENT_DICT[key], text)
        else:
            pos = cls.append(text)
            cls.CONTENT_DICT[key] = pos

    @classmethod
    def set_line(cls, pos, line, force=False):
        """Update a specific line."""
        lock = multiprocessing.Lock()
        lock.acquire()
        try:
            # Too short: add empty lines...
            while len(cls.CONTENT) <= pos:
                cls.CONTENT.append('')

            # print('CONTENT[{}] = {}'.format(pos,line), file=sys.stderr)
            cls.CONTENT[pos] = line

            if force:
                cls.redraw()
            # else:
            #     cls.time_refresh()

        finally:
            lock.release()

    @classmethod
    def append(cls, text):
        pos = len(cls.CONTENT)
        cls.set_line(pos, text)
        # cls.set_result('Now {} lines in CONTENT'.format(pos))

        if len(cls.CONTENT) - cls.SCROLL_POS > cls.LINES:
            cls.scroll_down()

        return pos

    @classmethod
    def __print_line(cls, pos, text, fmt=0, right=''):
        if cls.INITIATED:
            # lock = multiprocessing.Lock()
            # lock.acquire()
            try:
                # text = '{:3d}_{}'.format(pos,text)
                text = ' '+text
                right = right+' '
                prefix = '...'

                if cls.CONTENT_START <= pos < cls.CONTENT_START + cls.LINES:
                    # This is a content body line!
                    space = cls.WIDTH - cls.MAX_CONTENT_LENGTH - 2
                    if space >= 0:
                        text = text.ljust(cls.WIDTH)
                    else:
                        startpos = len(prefix) - space
                        text = (prefix + text[startpos:]).ljust(cls.WIDTH)
                else:
                    # This is a header or footer line!
                    space = cls.WIDTH - len(text) - len(right)
                    if space >= 0:
                        text = text + ' '*space + right
                    else:
                        startpos = len(prefix) - space
                        text = prefix + text[startpos:] + right

                # Make sure to only print in allowed area.
                if 0 <= pos < cls.HEIGHT:
                    try:
                        cls.WINDOW.move(0, 0)
                        # cls.WINDOW.refresh()
                        cls.WINDOW.addstr(pos, 0, text, fmt)
                        # cls.WINDOW.refresh()
                    except:
                        # print('pos={},text={},fmt={}'.format(pos,text,fmt),
                        # file=sys.stderr)
                        raise

            finally:
                # lock.release()
                pass

    @classmethod
    def __clear_line(cls, pos):
        if cls.INITIATED:
            cls.WINDOW.move(0, 0)
            # cls.WINDOW.refresh()
            cls.WINDOW.addstr(pos, 0, ' '*cls.WIDTH, 0)
            # cls.WINDOW.refresh()

    @classmethod
    def scroll_up(cls, lines=1):
        if cls.SCROLL_POS > 0:
            cls.SCROLL_POS -= lines
        else:
            cls.SCROLL_POS = 0
        cls.redraw_content()

    @classmethod
    def scroll_down(cls, lines=1):
        if len(cls.CONTENT) - cls.SCROLL_POS > cls.LINES:
            cls.SCROLL_POS += lines
        else:
            cls.SCROLL_POS = len(cls.CONTENT) - cls.LINES
        cls.redraw_content()

    @classmethod
    def print_status_line(cls):
        if cls.INITIATED:
            text = 'STATUS: ' + cls.STATUS
            if cls.PROGRESS is None:
                cls.__print_line(1, text, fmt=curses.A_REVERSE)
            else:
                cls.__print_line(1, text, fmt=curses.A_REVERSE,
                                 right=progressBar(*cls.PROGRESS))
        elif sys.stdout.isatty():
            if cls.PROGRESS is None:
                print('\r{}'.format(cls.STATUS), end='')
            else:
                print('\r{} {}'.format(cls.STATUS, progressBar(*cls.PROGRESS)),
                      end='')
            sys.stdout.flush()

    @classmethod
    def set_header(cls, text, force=True):
        cls.TITLE = text
        if force:
            cls.__print_line(0, cls.TITLE)

    @classmethod
    def set_status(cls, text, progress=None, force=True):
        cls.STATUS = text
        cls.PROGRESS = progress
        if force:
            cls.print_status_line()

    @classmethod
    def set_progress(cls, progress, total, force=True):
        cls.PROGRESS = (progress, total)
        if force:
            cls.print_status_line()

    @classmethod
    def set_result(cls, text, force=True):
        cls.RESULT = text
        if not cls.INITIATED and sys.stdout.isatty():
            print(cls.RESULT)
        elif force:
            cls.__print_line(cls.HEIGHT-3, cls.RESULT, fmt=curses.A_REVERSE)

    @classmethod
    def set_prompt(cls, text, force=True):
        cls.PROMPT = text
        if force:
            cls.__print_line(cls.HEIGHT-2, cls.PROMPT)

    @classmethod
    def wait_for_quit(cls):
        if cls.INITIATED:
            cls.redraw()
            cls.set_prompt('Press q to exit...')
            while 1:
                c = cls.WINDOW.getch()
                if c == curses.KEY_RESIZE:
                    cls.redraw()
                if c == ord('q'):
                    break
                elif c == curses.KEY_UP:
                    cls.scroll_up()
                elif c == curses.KEY_DOWN:
                    cls.scroll_down()
                elif c == curses.KEY_PPAGE:
                    cls.scroll_up(cls.LINES)
                elif c == curses.KEY_NPAGE:
                    cls.scroll_down(cls.LINES)
            cls.quit()

    @classmethod
    def quit(cls):
        """End curses application."""
        if cls.INITIATED:
            curses.nocbreak()
            curses.echo()
            curses.curs_set(1)
            curses.endwin()
            cls.INITIATED = False


# -----------------------------------------------------------------------------
def redraw():
    Screen.redraw()


def wait_for_quit():
    Screen.wait_for_quit()


def append(text):
    return Screen.append(text)


def update(key, text):
    return Screen.update(key, text)


# def separator(line):
#     printline(line, '='*Screen.WIDTH)


def header(text):
    Screen.set_header(text)


def status(text, progress=None):
    Screen.set_status(text, progress=progress)


def finish_status():
    if not Screen.INITIATED:
        print('')


def result(text):
    Screen.set_result(text)


def prompt(text):
    Screen.set_prompt(text)


def set_progress(progress, total, force=True):
    Screen.set_progress(progress, total, force=force)


# TERMINAL OUTPUT FORMATTING
# -----------------------------------------------------------------------------
PURPLE = '\033[95m'
CYAN = '\033[96m'
DARKCYAN = '\033[36m'
BLUE = '\033[94m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
RED = '\033[91m'
BOLD = '\033[1m'
UNDERLINE = '\033[4m'
END = '\033[0m'


def error(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, RED, message, END)
    else:
        return message


def warn(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, YELLOW, message, END)
    else:
        return message


def success(message):
    if False and sys.stdout.isatty() and not CONFIG['GENERAL']['USE_GUI']:
        return '{0}{1}{2}{3}'.format(BOLD, GREEN, message, END)
    else:
        return message



#!/usr/bin/env python
# coding=utf-8
""" This is the command line tool for the esahub python package.
"""
from __future__ import print_function
import locale
import warnings
import sys
import argparse
import time
import logging
import datetime
from .config import CONFIG


locale.setlocale(locale.LC_ALL, '')
warnings.filterwarnings("ignore",
                        message="WARNING: unsupported Compound type, skipping")
logger = logging.getLogger('esahub')


# -----------------------------------------------------------------------------
# PARSING COMMAND LINE OPTIONS
# -----------------------------------------------------------------------------
def parse_cli_options(args=None):
    """ Dedicated function for parsing command line options for esahub.
        An array of CLI arguments can be passed to simulate this behavior.
    """

    parser = argparse.ArgumentParser(
        description='Download satellite data archives.',
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    subparsers = parser.add_subparsers(help='sub-command help', dest='cmd')
    subparsers.required = True

    # Prepare different program streams:
    parser_get = subparsers.add_parser(
        'get', description='Download satellite data archives.',
        help='Download Sentinel data. \n  Type esahub get --help for details.',
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    parser_ls = subparsers.add_parser(
        'ls', description='Query satellite data archives for size.',
        help="""Investigate the amount of available data.\n"""
        """  Type esahub ls --help for details.""",
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    parser_doctor = subparsers.add_parser(
        'doctor', description='Check local data for corruption.',
        help="""Check already downloaded archives for consistency.\n"""
        """  Type esahub doctor --help for details.""",
        formatter_class=lambda prog: argparse.RawTextHelpFormatter(
                prog, max_help_position=40))

    # ARGUMENTS FOR ALL COMMANDS
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_ls, parser_doctor):
        p.add_argument('-N', '--nproc', type=int,
                       help='Number of simultaneous processes/downloads.')
        p.add_argument('--log', action='store_true',
                       help='Write a log file.')
        p.add_argument('--debug', action='store_true',
                       help='Enable debugging.')
        p.add_argument('--email', action='store_true',
                       help='Send an email report.')
        p.add_argument('--gui', action='store_true',
                       help='Use GUI.')

    # ARGUMENTS FOR GET AND DOCTOR
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_doctor):
        p.add_argument('-d', '--dir',
                       help='Specify the local data directory.')

    # ARGUMENTS FOR GET AND LS
    # -------------------------------------------------------------------------
    for p in (parser_get, parser_ls):
        p.add_argument(
            'sat', nargs='?', default=None,
            choices=list(CONFIG['SATELLITES'].keys()),
            help='Satellite')
        p.add_argument(
            '--mission',
            help="Select the mission (Sentinel-1, Sentinel-2, Sentinel-3).")
        p.add_argument(
            '--type',
            help="Select products matching specified product type (e.g. GRD).")
        p.add_argument(
            '--server', choices=list(CONFIG['SERVERS'].keys()),
            help="Specify the dowload server.")
        p.add_argument(
            '-g', '--geo', action='append',
            help="""Specify geospatial location for intersection as point """
                 """or polygon. You can add multiple options.\n"""
                 """  Examples: 'POINT(-6.26 53.35)' (Dublin) \n"""
                 """  'POLYGON((-25.1 46.8,-5.25 46.8,-5.25 57.4,"""
                 """-25.1 57.4,-25.1 46.8))' (Real Ireland)""")
        p.add_argument(
            '--location', action='append',
            choices=[l for l, value in CONFIG['LOCATIONS'].items()],
            help="""Specify location semantically as defined in config """
            """file. You can add multiple options.\n"""
            """  Examples: 'Ireland_Mace_Head'""")
        p.add_argument(
            '-t', '--time', choices=['today', 'yesterday', '24h', 'midnight'],
            help="""Shortcuts for specifying time intervals. """
                 """Options include:\n"""
                 """  today|yesterday|24h|midnight""")
        p.add_argument(
            '-A', '--from_time',
            help="""Select data after timestamp FROM. """
                 """Default is 1970-01-01T00:00:00.000Z.""")
        p.add_argument(
            '-B', '--to_time',
            help="Select data before timestamp TO. Default is NOW.")
        p.add_argument(
            '-q', '--query', help='Specify custom query.')

    # ARGUMENTS FOR GET ONLY
    # -------------------------------------------------------------------------
    for p in (parser_get,):
        p.add_argument(
            '--force', action='store_true',
            help='Force download even if index existing.')
        p.add_argument(
            '--in', help='Read download list from file.')

    # ARGUMENTS FOR LS ONLY
    # -------------------------------------------------------------------------
    for p in (parser_ls,):
        p.add_argument(
            '--out', help='Write list to file.')

    # ARGUMENTS FOR DOCTOR ONLY
    # -------------------------------------------------------------------------
    for p in (parser_doctor,):
        p.add_argument(
            '-m', '--mode',
            help="""Specify the mode for file checking. Options include:\n"""
                 """  file - check if archives are valid zip or """
                 """netcdf files (very fast).\n"""
                 """  md5 - check if archives match MD5 sum provided online """
                 """(can be slow).""")
        p.add_argument(
            '--repair', action='store_true',
            help='Redownload corrupt files.')
        p.add_argument(
            '--delete', action='store_true',
            help='Delete corrupt files. Ignored if running with --repair')

    if args is None:
        return vars(parser.parse_args())
    else:
        return vars(parser.parse_args(args))


# -----------------------------------------------------------------------------
def not_none(dictionary, key):
    return (key in dictionary and dictionary[key] is not None)


def set_config(args):
    global CONFIG
    if not_none(args, 'mode'):
        CONFIG['GENERAL']['CHECK_MODE'] = args['mode']
    if not_none(args, 'force'):
        CONFIG['GENERAL']['SKIP_EXISTING'] = not args['force']
    if not_none(args, 'nproc'):
        CONFIG['GENERAL']['N_DOWNLOADS'] = \
            CONFIG['GENERAL']['N_PROC'] = args['nproc']
    # if not_none(args, 'quiet'):
    #     CONFIG['GENERAL']['QUIET'] = args['quiet']
    if not_none(args, 'dir'):
        CONFIG['GENERAL']['TMP_DIR'] = args['dir']
    if not_none(args, 'target'):
        CONFIG['GENERAL']['DATA_DIR'] = args['target']
    if not_none(args, 'output'):
        CONFIG['GENERAL']['SOLR_INDEX_FILE'] = args['output']
    if not_none(args, 'email'):
        CONFIG['GENERAL']['SEND_EMAIL'] = args['email']
    if not_none(args, 'log'):
        CONFIG['GENERAL']['LOGGING'] = args['log']
    if not_none(args, 'gui'):
        CONFIG['GENERAL']['USE_GUI'] = args['gui']

    if not_none(args, 'sat'):
        CONFIG['GENERAL']['QUERY']['satellite'] = args['sat']
    if not_none(args, 'server'):
        CONFIG['GENERAL']['QUERY']['server'] = args['server']
    if not_none(args, 'mission'):
        CONFIG['GENERAL']['QUERY']['mission'] = args['mission']
    if not_none(args, 'geo'):
        CONFIG['GENERAL']['QUERY']['geo'] = args['geo']
    if not_none(args, 'location'):
        CONFIG['GENERAL']['QUERY']['location'] = args['location']
    if not_none(args, 'time'):
        CONFIG['GENERAL']['QUERY']['time'] = args['time']
    if not_none(args, 'from_time'):
        CONFIG['GENERAL']['QUERY']['from_time'] = args['from_time']
    if not_none(args, 'to_time'):
        CONFIG['GENERAL']['QUERY']['to_time'] = args['to_time']
    if not_none(args, 'type'):
        CONFIG['GENERAL']['QUERY']['type'] = args['type']
    if not_none(args, 'query'):
        CONFIG['GENERAL']['QUERY']['query'] = args['query']

    if not_none(args, 'in'):
        CONFIG['GENERAL']['IN_FILE'] = args['in']
    if not_none(args, 'out'):
        CONFIG['GENERAL']['OUT_FILE'] = args['out']


# -----------------------------------------------------------------------------
def shutdown():
    from esahub import tty
    msg = 'Execution interrupted manually.'
    logger.warning(msg)
    tty.quit()
    print(msg, file=sys.stderr)
    sys.exit()


def main():
    """ Execute selected command. """
    args = parse_cli_options()
    set_config(args)
    cmd = args['cmd']

    # This needs to happen before logging is enabled!
    CONFIG['GENERAL']['LOG_FILE'] = CONFIG['GENERAL']['LOG_FILE'].format(
        cmd=cmd,
        time=datetime.datetime.strftime(datetime.datetime.now(),
                                        '%Y-%m-%d_%H-%M-%S')
    )
    CONFIG['GENERAL']['EMAIL_SUBJECT'] = \
        CONFIG['GENERAL']['EMAIL_SUBJECT'].format(cmd=cmd)

    if CONFIG['GENERAL']['SEND_EMAIL']:
        CONFIG['GENERAL']['LOGGING'] = True

    if CONFIG['GENERAL']['LOGGING'] or args['debug']:
        logging.basicConfig(
            filename=CONFIG['GENERAL']['LOG_FILE'],
            format='[%(asctime)s] %(levelname)s - %(message)s',
            datefmt='%m/%d/%Y %H:%M:%S',
            level=logging.DEBUG if args['debug'] else logging.INFO
        )
    else:
        logging.disable(logging.CRITICAL)

    program = 'esahub {}'.format(' '.join(sys.argv[1:]))
    logger.info('Executing: {}'.format(program))

    #
    # These modules MUST be imported AFTER altering the CONFIG.
    #
    from esahub import tty, main

    try:
        a = time.time()

        tty.init()
        if CONFIG['GENERAL']['USE_GUI']:
            tty.header(program)

        if cmd == 'doctor':
            main.doctor(delete=args['delete'], repair=args['repair'])

        elif cmd == 'ls':
            main.ls()

        elif cmd == 'get':
            main.get()

    except KeyboardInterrupt:
        shutdown()

    finally:
        b = time.time()
        logger.info('TOTAL EXECUTION TIME: {0:.1f}s'.format(b-a))

    # Send email report.
    if CONFIG['GENERAL']['LOGGING'] and CONFIG['GENERAL']['SEND_EMAIL'] and \
            len(CONFIG['GENERAL']['EMAIL_REPORT_RECIPIENTS']):
        main.email()

    if CONFIG['GENERAL']['USE_GUI']:
        try:
            tty.wait_for_quit()
        except KeyboardInterrupt:
            shutdown()


if __name__ == '__main__':
    main()



# coding=utf-8
""" This module contains methods to check whether a local archive is corrupted.
    It can retrieve the true MD5 from SciHub and compare with the local file
    or check whether the archive is a valid zip file.
"""
from __future__ import print_function
from . import scihub, checksum, tty
from .config import CONFIG
import zipfile
import os
import sys
import logging
try:
    from netCDF4 import Dataset
    NETCDF_INSTALLED = True
except ImportError:
    NETCDF_INSTALLED = False

logger = logging.getLogger('esahub')
PY2 = sys.version_info < (3, 0)
try:
    ZIP_ERROR = zipfile.BadZipFile
except AttributeError:
    ZIP_ERROR = zipfile.BadZipfile


# CHECKING FILE CONSISTENY
# -----------------------------------------------------------------------------
FILE_COUNTER = 0
BAD_FILE_COUNTER = 0
BAD_FILES = []


# FILE LISTING
# -----------------------------------------------------------------------------

def local_ls():
    #
    # Collect a list of all files.
    #
    all_files = []
    for (root, subdirs, files) in os.walk(CONFIG['GENERAL']['TMP_DIR']):
        for f in files:
            full_file_path = os.path.join(root, f)
            if f.endswith('.zip') or f.endswith('.nc'):
                all_files.append(full_file_path)
    return all_files


def check_file(full_file_path, mode):
    """ Check an already downloaded file for consistency.

    Parameters
    ----------
    full_file_path : str
    mode : {'file', 'md5'}
        If `md5`, check if the md5 checksum matches the value stored on SciHub.
        This is safer, but potentially slow. Also, the md5 checksum may not be
        available for older files. If `file`, check if the archive is a valid
        zip archive or a valid netCDF file. This method is very fast and should
        be accurate in the case of interrupted downloads.

    Returns
    -------
    tuple (str, bool, str)
        A tuple containing the file path, a boolean whether or not the file is
        okay, and a status message.
    """
    message = None

    if mode == 'file':
        #
        # Check if the archive is a valid zip archive or
        # a valid netCDF file
        #
        ext = os.path.splitext(full_file_path)[1]
        if ext == '.zip':
            try:
                zip_ref = zipfile.ZipFile(full_file_path, 'r')
                zip_ref.close()
            except ZIP_ERROR as e:
                message = '{} {}'.format(tty.error('BAD ZIP FILE:'),
                                         full_file_path)
                healthy = False
            else:
                message = '{} {}'.format(tty.success('OKAY:'), full_file_path)
                healthy = True
        elif ext == '.nc':
            if not NETCDF_INSTALLED:
                raise ImportError("Need to install `netCDF4` to use this "
                                  "feature!")
            try:
                nc_ref = Dataset(full_file_path, 'r')
                nc_ref.close()
            except (OSError, IOError) as e:
                message = '{} {}'.format(tty.error('BAD NETCDF FILE:'),
                                         full_file_path)
                healthy = False
            else:
                message = '{} {}'.format(tty.success('OKAY:'), full_file_path)
                healthy = True
        else:
            message = '{} {}'.format(tty.error('UNKNOWN FILE FORMAT:'),
                                     full_file_path)
            healthy = False

    elif mode == 'md5':
        #
        # Check the md5sum against SciHub.
        #
        remote_md5 = scihub.md5(product=full_file_path)
        if remote_md5 is False:
            message = '{} {}'.format(tty.error('MD5 NOT FOUND:'),
                                     full_file_path)
            healthy = False
        else:
            if checksum.md5(full_file_path) == remote_md5:
                message = '{} {}'.format(tty.success('MD5 OKAY:'),
                                         full_file_path)
                healthy = True
            else:
                message = '{} {}'.format(tty.error('FILE CORRUPT:'),
                                         full_file_path)
                healthy = False

    return (full_file_path, healthy, message)


def _init_bad_file_counter():
    global FILE_COUNTER
    global BAD_FILE_COUNTER
    global BAD_FILES
    FILE_COUNTER = 0
    BAD_FILE_COUNTER = 0
    BAD_FILES = []


def _register_bad_file(status):
    file_path, healthy, message = status
    global FILE_COUNTER
    global BAD_FILE_COUNTER
    global BAD_FILES
    FILE_COUNTER += 1
    if not healthy:
        BAD_FILE_COUNTER += 1
        BAD_FILES.append(file_path)

    tty.update(os.path.split(file_path)[1],
               {'msg': message})
    logging.info(message)



# import distutils.cmd
import setuptools
import setuptools.command.build_py
import os
import shutil


def init_config():
    src = 'esahub/config.yaml'
    dst = os.path.expanduser('~/.esahub.conf')
    if not os.path.isfile(dst):
        shutil.copyfile(src, dst)


class BuildPyCommand(setuptools.command.build_py.build_py):
    """Custom build command."""

    def run(self):
        init_config()
        setuptools.command.build_py.build_py.run(self)


setuptools.setup(
    use_scm_version=True,
    cmdclass={
        'build_py': BuildPyCommand,
    }
)



# coding=utf-8
""" This module contains the functionality to extract the metadata
    from a Satellite archive, construct from that the relevant
    Solr index and post the index to a running Solr instance.
"""
from __future__ import print_function
import glob, sys, os, time, fnmatch, traceback, shutil
import re
import gc
import datetime as DT
import zipfile, tarfile
import multiprocessing
import subprocess
import lxml.etree as ET
import logginßg
import tempfile
from functools import partial
import pysolr
import json
from netCDF4 import Dataset
import numpy as np
from . import geo, helpers, tty, checksum
from .config import CONFIG

logger = logging.getLogger('evdc')
PY2 = sys.version_info < (3, 0)

if PY2:
    from StringIO import StringIO
else:
    from io import StringIO, BytesIO


# -----------------------------------------------------------------------------
# -----------------------------------------------------------------------------

def exists(product, where='solr'):
    """Check if a product is already indexed.

    Parameters
    ----------
    product : str
        Filename of the product to be indexed.
    where : str, optional
        If 'solr', check if the product exists on the Solr index.
        If 'local', check if the extracted metadata exists in the local data
        directory. (default: 'solr')

    Returns
    -------
    bool
        True if the product is already indexed, False otherwise.
    """
    product = os.path.split(product)[1]
    if where == 'solr':
        return (count(q='product_name:{}*'.format(
                    os.path.splitext(product)[0])) > 0)
    elif where == 'local':
        stripped_product_path = os.path.join(CONFIG['GENERAL']['DATA_DIR'],
                                             product)
        if os.path.splitext(stripped_product_path)[1] != '.zip':
            stripped_product_path += '.zip'
        return os.path.isfile(stripped_product_path)


def index(product, is_data=True, store=False, queue=None):
    """Extract the metadata from a single satellite product.

    Parameters
    ----------
    product : str
        Filename of the product to be indexed.
    is_data : bool, optional
        If True, compute md5 sum and file size from the file. Ingestiondate is
        current timestamp. Otherwise, get info from S3 storage.
    store : bool, optional
        If True, store the stripped metadata in
        `config.CONFIG['GENERAL']['DATA_DIR']` (default: False).
    queue : multiprocessing.Manager.Queue, optional
        A multiprocessing queue to submit status messages.

    Returns
    -------
    dict
        If successful, returns a dictionary of the extracted metadata.
        Otherwise returns None.
    """
    product_path, product_file = os.path.split(product)
    identifier = os.path.splitext(product_file)[0]
    satellite = helpers.get_satellite(product)

    try:
        if satellite in ['S1A', 'S1B', 'S2A', 'S2B', 'S3A']:
            DATA = _index_S123(product)
        elif satellite == 'S5P':
            DATA = _index_S5(product)
        elif satellite == 'AE':
            DATA = _index_AE(product)
        else:
            return {}

    except Exception as e:
        msg = 'Could not extract metadata from {}: {}'.format(product, e)
        if queue is None:
            tty.update(product_file, msg)
        else:
            queue.put((product_file,msg))
        logging.info(msg)
        traceback.print_exc(file=sys.stderr)
        return {}

    #
    # Add a number of custom properties.
    #

    # Compute the area covered.
    if 'footprint' in DATA:
        # DATA['area_sqkm'] = '{:.2f}'.format(geo.polygon_area(DATA['footprint']))
        DATA['area_sqkm'] = geo.polygon_area(DATA['footprint'])

    if is_data:
        #
        # If the file is the actual product, compute additional metadata such
        # as the md5 sum and the file size from the file itself.
        #
        DATA['inventory_time'] = DT.datetime.strftime(DT.datetime.now(),
                                                      '%Y-%m-%dT%H:%M:%SZ')
        # DATA['inventory_time'] = DT.datetime.now()
        # Add the md5 checksum to metadata
        DATA['md5'] = checksum.md5(product)
        # Add the file size to metadata
        DATA['bytes'] = os.path.getsize(product)

    else:
        #
        # If the file is only the stripped metadata, obtain additional metadata
        # from other source.
        #
        pass
        # if not 'md5' in DATA or not 'bytes' in DATA or not 'inventory_time' in DATA:
        #     s3_info = storage.s3_ls(identifier)
        #     if len(s3_info) == 1:
        #         DATA['inventory_time'] = DT.datetime.strftime(s3_info[0]['time'],'%Y-%m-%dT%H:%M:%SZ')
        #         # DATA['inventory_time'] = s3_info[0]['time']
        #         DATA['md5'] = s3_info[0]['md5']
        #         DATA['bytes'] = s3_info[0]['size']

    # Store a stripped copy
    if store:
        evdc_meta = {_: DATA[_] for _ in ['md5', 'bytes', 'inventory_time']}
        strip(product, target=CONFIG['GENERAL']['DATA_DIR'], data=evdc_meta)

    return DATA


def _index_S123(product):
    """Extract the metadata from a Sentinel-1,2,3 product.

    Parameters
    ----------
    product : str
        The full file path of a Sentinel-1,2,3 product (zip-archive).

    Returns
    -------
    dict
        A dictionary of the extracted metadata.
    """
    product_path, product_file = os.path.split(product)
    DATA = {}
    #
    # Decide which satellite the zip belongs to:
    # S1, S2, S3
    #
    satellite = helpers.get_satellite(product_file)
    suffix = CONFIG['SATELLITES'][satellite]['suffix']
    identifier = os.path.splitext(product_file)[0]
    with zipfile.ZipFile(product, 'r') as zip_ref:
        for meta_data_file, meta in CONFIG['META'][CONFIG['SATELLITES'][satellite]['meta']].items():
            zpath = os.path.join(identifier+suffix, meta_data_file)
            # Treat the specified filename as a regular expression.
            zfiles = [zfile for zfile in zip_ref.namelist() if re.search(zpath,zfile)]
            for zfile in zfiles:
                root = ET.fromstring( zip_ref.read(zfile) )
                pre = CONFIG['META_NAMESPACES'].copy()
                pre.update(root.nsmap)

                #
                # Some satellite specific things
                #
                if satellite == 'S3A':
                    INSTR = root.find("./metadataSection/metadataObject[@ID='platform']//sentinel-safe:instrument/sentinel-safe:familyName", pre).attrib['abbreviation'].lower()
                else:
                    INSTR = ''

                #
                # Loop through the specified meta data entries for the given file.
                # Find the entry in the document and append the property to the
                # DATA dictionary.
                #
                for key,item in meta.items():
                    try:
                        xml_path = item[0].format(INSTR=INSTR)
                        if item[1] == 'text':
                            val = root.find(xml_path, pre).text
                        else:
                            val = root.find(xml_path, pre).attrib[item[1]]
                        if len(item) == 3:
                            conversion_fn = eval(item[2])
                            val = conversion_fn(val)
                    except Exception as e:
                        #
                        # This entry doesn't exist. Move on.
                        #
                        pass
                    else:
                        DATA[key] = val
        #
        # Extract additional metadata stored with the file
        # (e.g. during dat stripping).
        #
        root = os.path.commonprefix(zip_ref.namelist())
        evdc_meta_file = os.path.join(root, '.evdc_meta')
        if evdc_meta_file in zip_ref.namelist():
            json_str = zip_ref.read(evdc_meta_file)
            if not PY2:
                json_str = json_str.decode()
            evdc_meta = json.loads(json_str)
            DATA.update(evdc_meta)

    #
    # On Sentinel-2, the manifest.safe file specifies familyName as SENTINEL
    # Since we want to be able to query for mission, make this more specific.
    #
    if satellite.startswith('S1'):
        DATA['family_name'] = 'Sentinel-1'
    elif satellite.startswith('S2'):
        DATA['family_name'] = 'Sentinel-2'

    DATA['product_name'] = identifier + suffix

    return DATA


def _index_AE(product):
    """Extract the metadata from an ADM-AEOLUS product.

    Parameters
    ----------
    product : str
        The full file path of an ADM-AEOLUS product (tgz-archive).

    Returns
    -------
    dict
        A dictionary of the extracted metadata.
    """
    product_path, product_file = os.path.split(product)
    identifier = os.path.splitext(product_file)[0]
    satellite = 'AE'
    suffix = CONFIG['SATELLITES'][satellite]['suffix']
    DATA = {}
    with tarfile.open(product, 'r') as tar_ref:
        rootdir = os.path.commonprefix([os.path.split(_)[0] for _ in tar_ref.getnames()])
        for meta_data_file, meta in CONFIG['META'][CONFIG['SATELLITES'][satellite]['meta']].items():
            tpath = os.path.join(rootdir, meta_data_file)
            # Treat the specified filename as a regular expression.
            tfiles = [tfile for tfile in tar_ref.getnames() if re.search(tpath,tfile) and not os.path.split(tfile)[1].startswith('.')]
            for tfile in tfiles:
                tfile_object = tar_ref.extractfile(tfile)
                root = ET.fromstring( tfile_object.read() )
                tfile_object.close()
                pre = CONFIG['META_NAMESPACES'].copy()
                pre.update(root.nsmap)

                #
                # Loop through the specified meta data entries for the given file.
                # Find the entry in the document and append the property to the
                # DATA dictionary.
                #
                for key,item in meta.items():
                    try:
                        xml_path = item[0]
                        if item[1] == 'text':
                            val = root.find(xml_path, pre).text
                        else:
                            val = root.find(xml_path, pre).attrib[item[1]]
                        if len(item) == 3:
                            conversion_fn = eval(item[2])
                            val = conversion_fn(val)
                    except Exception as e:
                        #
                        # This entry doesn't exist. Move on.
                        #
                        pass
                    else:
                        DATA[key] = val
        #
        # Extract additional metadata stored with the file
        # (e.g. during data stripping).
        #
        evdc_meta_file = os.path.join(rootdir, '.evdc_meta')
        if evdc_meta_file in tar_ref.getnames():
            fobj = tar_ref.extractfile(evdc_meta_file)
            json_str = fobj.read()
            fobj.close()
            if not PY2:
                json_str = json_str.decode()
            evdc_meta = json.loads(json_str)
            DATA.update(evdc_meta)

    DATA['product_name'] = identifier + suffix

    if 'latstart' in DATA and 'lonstart' in DATA and 'latstop' in DATA and 'lonstop' in DATA:
        ymin=int(DATA['latstart'])/1.e6
        xmin=int(DATA['lonstart'])/1.e6
        ymax=int(DATA['latstop'])/1.e6
        xmax=int(DATA['lonstop'])/1.e6
        if xmax > 180: xmax -= 360
        if xmin > 180: xmin -= 360
        DATA['footprint'] = 'POLYGON(({xmin} {ymin},{xmin} {ymax},{xmax} {ymax},{xmax} {ymin},{xmin} {ymin}))'.format(
            ymin=ymin, xmin=xmin, ymax=ymax, xmax=xmax
        )
        del DATA['latstart'], DATA['lonstart'], DATA['latstop'], DATA['lonstop']

    return DATA


def ncget(nc_fid, path):
    """Return metadata attribute from netCDF file.

    Parameters
    ----------
    nc_fid : netCDF4.Dataset
        An opened netCDF Dataset.
    path : str
        The path to the metadata attribute. Consists of nested groups and the attribute separated
        by slashes:
        `group/subgroup/.../[attribute]`

    Returns
    -------
    str
        The value of the requested attribute. If no attribute is specified (i.e. the path ends
        with a slash), the last subgroup is returned.
    """
    group = nc_fid
    path_elements = path.split('/')
    groups = path_elements[:-1]
    attr = path_elements[-1]
    for _ in groups: group = group.groups[_]
    if len(attr) == 0:
        return group
    else:
        value = group.getncattr(attr)
        if isinstance(value, np.integer):
            # There are problems with JSON-serializing numpy integers,
            # therefore convert to native int:
            value = int(value)
        return value


def _index_S5(product):
    """Extract the metadata from a Sentinel-5P product.

    Parameters
    ----------
    product : str
        The full file path of a Sentinel-5P product (netCDF file).

    Returns
    -------
    dict
        A dictionary of the extracted metadata.
    """
    product_path, product_file = os.path.split(product)
    DATA = {}
    satellite = product_file[:3]

    # If the product is a zip archive, extract to temporary directory
    if os.path.splitext(product_file)[1] == '.zip':
        compressed = True
        tmpdir = tempfile.mkdtemp()
        z = zipfile.ZipFile(product,'r')
        product = z.extract(z.namelist()[0], tmpdir)
    else:
        compressed = False

    nc_fid = Dataset(product, 'r')

    for key,item in CONFIG['META'][CONFIG['SATELLITES'][satellite]['meta']].items():
        try:
            if len(item)==1:
                DATA[key] = ncget(nc_fid, item[0])
            elif len(item)==2:
                conversion_fn = eval(item[1])
                DATA[key] = conversion_fn(ncget(nc_fid, item[0]))
        except Exception as e:
            pass

    #
    # Extract additional metadata stored with the file
    # (e.g. during dat stripping).
    #
    if 'EVDC_META' in nc_fid.groups:
        meta_group = nc_fid.groups['EVDC_META']
        evdc_meta = {_ : ncget(meta_group,_) for _ in meta_group.ncattrs()}
        DATA.update(evdc_meta)

    nc_fid.close()
    gc.collect()

    if compressed: shutil.rmtree(tmpdir)

    return DATA


def strip(product, target, compress=True, data=None):
    """Create a stripped down copy of the satellite product that only contains the metadata.

    Parameters
    ----------
    product : str
        The full file path of a Sentinel product.
    target : str, optional
        The directory to store the stripped down product.
    compress : bool, optional
        If True, compress bare netCDF files in a zip archive (default: True)
    data : dict, optional
        Additional metadata to be stored. This will typically be the md5 checksum and filesize.

    Returns
    -------
    str
        If successful, the file path of the stripped copy. False otherwise.
    """
    satellite = helpers.get_satellite(product)
    if satellite in ['S1A', 'S1B', 'S2A', 'S2B', 'S3A']:
        return _strip_S123(product, target, data=data)
    elif satellite == 'S5P':
        return _strip_S5(product, target, compress, data=data)
    elif satellite == 'AE':
        return _strip_AE(product, target, data=data)
    else:
        return False


def _strip_S123(product, target, data):
    """Create a stripped down copy of the Sentinel-1,2,3 product that only contains the metadata.

    Parameters
    ----------
    product : str
        The full file path of a Sentinel-1,2,3 product.
    target : str, optional
        The directory to store the stripped down product.
    data : dict, optional
        Additional metadata to be stored. This will typically be the md5 checksum and filesize.

    Returns
    -------
    str
        If successful, the file path of the stripped copy. False otherwise.
    """

    product_path, product_file = os.path.split(product)
    satellite = helpers.get_satellite(product)
    stripped_product_path = os.path.join( target, product_file )

    with zipfile.ZipFile(stripped_product_path, 'w', compression=zipfile.ZIP_DEFLATED) as zip_stripped:
        with zipfile.ZipFile(product, 'r') as zip_original:
            root = os.path.commonprefix(zip_original.namelist())
            for meta_data_file in CONFIG['META'][CONFIG['SATELLITES'][satellite]['meta']]:
                zpath = os.path.join(root, meta_data_file)
                # Treat the specified filename as a regular expression.
                zfiles = [zfile for zfile in zip_original.namelist() if re.search(zpath,zfile)]
                for zfile in zfiles:
                    # Add file to new zip archive
                    zip_stripped.writestr(zfile, zip_original.read(zfile))

            #
            # Add additional metadata to zip archive
            #
            if data is not None:
                jsondata = json.dumps(data)
                zip_stripped.writestr(os.path.join(root, '.evdc_meta'), jsondata)

    return stripped_product_path


def _strip_AE(product, target, data):
    """Create a stripped down copy of the ADM-AEOLUS product that only contains the metadata.

    Parameters
    ----------
    product : str
        The full file path of an ADM-AEOLUS product.
    target : str, optional
        The directory to store the stripped down product.
    data : dict, optional
        Additional metadata to be stored. This will typically be the md5 checksum and filesize.

    Returns
    -------
    str
        If successful, the file path of the stripped copy. False otherwise.
    """

    product_path, product_file = os.path.split(product)
    satellite = helpers.get_satellite(product)
    stripped_product_path = os.path.join( target, product_file )

    with tarfile.open(stripped_product_path, 'w') as tar_stripped:
        with tarfile.open(product, 'r') as tar_original:
            root = os.path.commonprefix(tar_original.getnames())
            for meta_data_file in CONFIG['META'][CONFIG['SATELLITES'][satellite]['meta']]:
                tpath = os.path.join(root, meta_data_file)
                # Treat the specified filename as a regular expression.
                tfiles = [tfile for tfile in tar_original.getnames() if re.search(tpath,tfile) and not os.path.split(tfile)[1].startswith('.')]
                for tfile in tfiles:
                    # Add file to new tar archive
                    fobj = tar_original.extractfile(tfile)
                    tar_stripped.addfile(tar_original.getmember(tfile), fobj)
                    fobj.close()

            #
            # Add additional metadata to tar archive
            #
            if data is not None:
                jsondata = json.dumps(data).encode('utf-8')
                tfile = os.path.join(root, '.evdc_meta')
                tinfo = tarfile.TarInfo(tfile)
                fobj = BytesIO(jsondata)
                tinfo.size = len(jsondata)
                tar_stripped.addfile(tinfo, fobj)

    return stripped_product_path


def _strip_S5(product, target, compress, data=None):
    """Create a stripped down copy of the Sentinel-5 product that only contains the metadata.

    Parameters
    ----------
    product : str
        The full file path of a Sentinel-5 product.
    target : str
        The directory to store the stripped down product.
    compress : bool
        If True, compress the stripped product in a zip archive.
    data : dict, optional
        Additional metadata to be stored. This will typically be the md5 checksum and filesize.

    Returns
    -------
    str
        If successful, the file path of the (compressed) stripped copy. False otherwise.
    """
    product_path, product_file = os.path.split(product)

    if compress:
        stripped_product_path = tempfile.mkstemp()[1]
    else:
        stripped_product_path = os.path.join( target, product_file )

    try:
        nc_fid = Dataset(product, 'r')
        nc_stripped = Dataset(stripped_product_path, 'w')
    except:
        print(product)
        traceback.print_exc(file=sys.stderr)
        return False

    def _copy(src,tgt):
        # Set attributes
        tgt.setncatts({_:src.getncattr(_) for _ in src.ncattrs()})
        # Recursively go through groups
        for src_group_name,src_group in src.groups.items():
            tgt_group = tgt.createGroup(src_group_name)
            _copy(src_group, tgt_group)

    _copy(nc_fid, nc_stripped)

    #
    # Add additional metadata to zip archive
    #
    if data is not None:
        meta_group = nc_stripped.createGroup('EVDC_META')
        meta_group.setncatts(data)

    nc_stripped.close()

    if compress:
        zipped_stripped_product_path = os.path.join( target, product_file + '.zip' )
        with zipfile.ZipFile(zipped_stripped_product_path, 'w', compression=zipfile.ZIP_DEFLATED) as zip_stripped:
            zip_stripped.write(stripped_product_path)
        os.remove(stripped_product_path)
        return zipped_stripped_product_path
    else:
        return stripped_product_path


COUNT = 0
def _count(boolean,total=None):
    global COUNT
    if boolean: COUNT += 1
    if total is not None: tty.set_progress(COUNT,total)

COLLECT = []


def _init_collect():
    global COLLECT, COUNT
    COLLECT = []
    COUNT = 0


def _collect(result,total=None):
    global COLLECT, COUNT
    COLLECT.append(result)
    COUNT += 1
    if total is not None: tty.set_progress(COUNT,total)


def index_all(directory, is_data=True, store=False, daily=False, subset=None):
    """Generate the index for all satellite products in the data directory.

    Parameters
    ----------
    directory : str
        The directory containing the products to be indexed.
    is_data : bool, optional
        If True, compute md5 sum and file size from the file. Otherwise, get info from S3 storage.
    store : bool, optional
        If True, store the stripped metadata in `config.CONFIG['GENERAL']['DATA_DIR']`
        (default: False).
    daily : bool, optional
        Whether to only index products that were added since yesterday.
    subset : list of str, optional
        A list of filenames to be indexed.

    Returns
    -------
    list of dict
        A list of the metadata dictonary of each satellite product in the directory.
    """
    start_time = time.time()

    #
    # Get a list of all products in directory.
    #
    products = helpers.ls(directory, subset=subset)

    if daily:
        # Only index products that were modified after yesterday midnight.
        yesterday = DT.datetime.combine(DT.date.today() - DT.timedelta(1), DT.datetime.min.time())
        products = [_ for _ in products if DT.datetime.fromtimestamp(os.path.getmtime(_)) > yesterday]

    #
    # Index all products found (multiprocessing).
    #
    msg = 'Indexing {:d} products... '.format(len(products))
    tty.status(msg)

    q = multiprocessing.Manager().Queue()
    pool = multiprocessing.Pool(processes=CONFIG['GENERAL']['N_PROC'])

    _init_collect()
    _collect_metadata = partial(_collect, total=len(products))
    _index_fn = partial(index, is_data=is_data, store=store)

    for product in products:
        pool.apply_async(_index_fn, (product,), callback=_collect_metadata)

    pool.close()
    pool.join()

    stop_time = time.time()

    msg = 'Indexing {:d} products... ({:.1f}s)'.format(len(products), stop_time-start_time)
    logging.info(msg)
    tty.status(msg)

    return COLLECT


def commit(data, quiet=False):
    """Feeds data into the running Solr instance.

    Parameters
    ----------
    data : dict or list of dict
        The data is directly piped into Solr.
    quiet : bool, optional
        Whether to suppress output.

    Returns
    -------
    str
        The Solr output.
    """
    solr = pysolr.Solr( os.path.join(CONFIG['GENERAL']['SOLR_URL'], CONFIG['GENERAL']['SOLR_CORE']) )

    if type(data) is dict: data = [data]

    if not quiet:
        msg = 'Posting {} documents to Solr core {}...'.format(len(data), CONFIG['GENERAL']['SOLR_CORE'])
        tty.status(msg)
    start_time = time.time()
    try:
        result = None
        for chunk in helpers.chunks(data, 1000):
            result = solr.add(chunk)
    except Exception as e:
        logging.error('Could not add documents to Solr: {}'.format(e))
        return e
    else:
        stop_time = time.time()
        if not quiet:
            msg = 'Posting {} documents to Solr core {}... done. ({:.1f}s)'.format(len(data), CONFIG['GENERAL']['SOLR_CORE'], stop_time-start_time)
            logging.info(msg)
            tty.status(msg)
    finally:
        solr.get_session().close()

    return result


def _wipe():
    """Wipes the entire index. Use with caution."""
    solr = pysolr.Solr( os.path.join(CONFIG['GENERAL']['SOLR_URL'], CONFIG['GENERAL']['SOLR_CORE']), timeout=10 )
    solr.delete(q='*:*')
    solr.get_session().close()


def count(q='*:*'):
    """Returns the number of documents in the Solr index matching a query.

    Parameters
    ----------
    q : str, optional
        Solr query string (default: '*:*').

    Returns
    -------
    int
        The number of documents found.
    """
    solr = pysolr.Solr( os.path.join(CONFIG['GENERAL']['SOLR_URL'], CONFIG['GENERAL']['SOLR_CORE']), timeout=10 )
    result = solr.search(q)
    solr.get_session().close()
    return result.hits


def search(*args, **kwargs):
    """Returns documents from the Solr index. Follows the Syntax of pysolr.Solr.search()."""
    solr = pysolr.Solr( os.path.join(CONFIG['GENERAL']['SOLR_URL'], CONFIG['GENERAL']['SOLR_CORE']), timeout=10 )
    result = solr.search(*args, **kwargs)
    solr.get_session().close()
    return result
